#!/usr/bin/env python

import os
import sys
import json
import time
import h5py
import pickle
import logging
import argparse
import warnings
import numpy as np
import pandas as pd
import Bio.Seq

from functools import lru_cache

if __name__ != '__main__':
    from ._version import __version__
    #import inStrain.profileUtilities

#import inStrain.profileUtilities
import inStrain.profile.profile_utilities
import inStrain.logUtils
import inStrain.filter_reads

class SNVprofile:
    '''
    This object holds the profile of a single .fasta / .bam pair
    '''

    firstLevels = ['output', 'raw_data', 'log', 'figures']

    def __init__(self, location):
        # Parse and store the location
        self.location = os.path.abspath(location)

        # Make the file structure
        self._make_fileStructure()

        # Initialize attributes
        self._initialize_attributes()

    def store(self, name, value, type, description):
        '''
        Store an attribute in the attributes.tsv file

        Args:
            name = name of attribute
            value = actual value of the attribute
            type = type of attribute
            description = one-line description of the attribute
        '''
        # Prepare for storage
        if type == 'value':
            adb = pd.DataFrame({'value':value, 'type':type, 'description':description}, index=[name])

        elif type == 'dictionary':
            fileloc = self._get_fileloc(name) + '.json'
            self._store_dictionary(value, fileloc)
            adb = pd.DataFrame({'value':fileloc, 'type':type, 'description':description}, index=[name])

        elif type == 'list':
            fileloc = self._get_fileloc(name) + '.txt'
            self._store_list(value, fileloc)
            adb = pd.DataFrame({'value':fileloc, 'type':type, 'description':description}, index=[name])

        elif type == 'numpy':
            fileloc = self._get_fileloc(name) + '.npz'
            self._store_numpy(value, fileloc)
            adb = pd.DataFrame({'value':fileloc, 'type':type, 'description':description}, index=[name])

        elif type == 'pandas':
            fileloc = self._get_fileloc(name) + '.csv.gz'
            self._store_pandas(value, fileloc)
            adb = pd.DataFrame({'value':fileloc, 'type':type, 'description':description}, index=[name])

        elif type == 'special':
            fileloc = self._store_special(value, name)
            adb = pd.DataFrame({'value':fileloc, 'type':type, 'description':description}, index=[name])

        elif type == 'pickle':
            fileloc = self._get_fileloc(name) + '.pickle'
            self._store_pickle(value, fileloc)
            adb = pd.DataFrame({'value':fileloc, 'type':type, 'description':description}, index=[name])

        else:
            print(logging.error('I dont know how to save a {0} type, so Im just going to pickle it'.format(type)))
            fileloc = self._get_fileloc(name) + '.pickle'
            self._store_pickle(value, fileloc)
            adb = pd.DataFrame({'value':fileloc, 'type':'pickle', 'description':description}, index=[name])

        # See if it's already in there
        Adb = self._get_attributes_file()
        if name in Adb.index:

            # Make sure it's the same type and description; otherwise crash
            for thing in ['type', 'description']:
                if (Adb.loc[name, thing] != eval(thing)):
                    logging.error('WILL NOT OVERWRITE {0}; {1} arent the same ({2} vs {3}))'.format(
                            name, thing, Adb.loc[name, thing], eval(thing)))
                    return

            # Overwrite
            Adb.at[name, 'value'] = adb['value'].tolist()[0]

        # Add the value if not
        else:
            Adb = Adb.append(adb)

        self._store_attributes_file(Adb)

    def get(self, name, **kwargs):
        '''
        Get an attribute from the attributes.tsv file
        '''
        Adb = self._get_attributes_file()
        if name not in Adb.index:
            #logging.error("{0} is not in this IS object".format(name))
            return None

        type = Adb.loc[name, 'type']

        if type == 'value':
            return Adb.loc[name, 'value']

        filename = self.get_location('raw_data') + '/' + os.path.basename(Adb.loc[name, 'value'])

        if type == 'dictionary':
            return self._load_dictionary(filename)

        elif type == 'list':
            return self._load_list(filename)

        elif type == 'numpy':
            return self._load_numpy(filename)

        elif type == 'pandas':
            return self._load_pandas(filename)

        elif type == 'special':
            return self._load_special(filename, name, **kwargs)

        elif type == 'pickle':
            return self._load_pickle(filename)

        else:
            logging.error('I dont know how to load a {0} type!'.format(type))

    def get_location(self, name):
        if name == 'output':
            return os.path.join(self.location, 'output/')
        if name == 'raw_data':
            return os.path.join(self.location, 'raw_data/')
        if name == 'log':
            return os.path.join(self.location, 'log/')
        if name == 'figures':
            # To maintain backwards compatibility
            loc = os.path.join(self.location, 'figures/')
            if not os.path.exists(loc):
                os.makedirs(loc)
            return loc
        else:
            logging.error('I dont know the location of {0}!'.format(name))

    def generate(self, name, store=True, return_table=False, **kwargs):
        '''
        Generate user-facing output tables based on information stored in the IS
        '''

        report_mm_level = kwargs.get('mm_level', False)
        if name == 'SNVs':
            column_order = ['scaffold', 'position', 'position_coverage', 'allele_count',
                            'ref_base', 'con_base', 'var_base',
                            'ref_freq', 'con_freq', 'var_freq',
                            'A', 'C', 'T', 'G',
                            'gene', 'mutation', 'mutation_type', 'cryptic']

            # Get the base table
            db = self.get_nonredundant_snv_table()

            # Do you have the mutation types? If so, merge in
            mdb = self.get('SNP_mutation_types')
            if (mdb is not None):
                if len(mdb) > 0:
                    mdb = mdb[['scaffold', 'position', 'mutation_type', 'mutation', 'gene']]
                    db = pd.merge(db, mdb, how='left', on=['scaffold', 'position'])

            # Do you have information about coverage? If so, merge in
            pass

            db = reorder_columns(db, column_order)

        elif name == 'scaffold_info':
            column_order = ['scaffold', 'length', 'coverage', 'breadth',
                            'nucl_diversity',
                            'coverage_median', 'coverage_std', 'coverage_SEM',
                            'breadth_minCov', 'breadth_expected',
                            'nucl_diversity_median',
                            'nucl_diversity_rarefied', 'nucl_diversity_rarefied_median',
                            'breadth_rarefied',
                            'conANI_reference', 'popANI_reference',
                            'SNS_count', 'SNV_count', 'divergent_site_count']

            db = self.get_nonredundant_scaffold_table()
            db = reorder_columns(db, column_order)

        elif name == 'linkage':
            column_order = ['scaffold', 'position_A', 'position_B', 'distance',
                            'r2', 'd_prime',
                            'r2_normalized', 'd_prime_normalized',
                            'allele_A', 'allele_a',
                            'allele_B', 'allele_b',
                            'countab', 'countAb', 'countaB', 'countAB', 'total']
            db = self.get_nonredundant_linkage_table()
            db = reorder_columns(db, column_order)

        elif name == 'gene_info':
            column_order = ['scaffold', 'gene', 'gene_length',
                            'coverage', 'breadth', 'breadth_minCov', 'nucl_diversity',
                             'start', 'end', 'direction', 'partial',
                             'dNdS_substitutions', 'pNpS_variants',
                             'SNV_count', 'SNV_S_count', 'SNV_N_count',
                             'SNS_count', 'SNS_S_count', 'SNS_N_count',
                             'divergent_site_count']

            Gdb = self.get('genes_table')
            if Gdb is None:
                logging.info("Cannot generate genes_table, no genes were profiled")
                return

            for thing in ['genes_coverage', 'genes_clonality', 'genes_SNP_count']:
                db = self.get(thing)
                if db is None:
                    logging.debug('Skipping {0} gene calculation; you have none'.format(thing))
                    continue
                if len(db) == 0:
                    logging.debug('Skipping {0} gene calculation; you have none'.format(thing))
                    continue
                db = db.sort_values('mm').drop_duplicates(subset=['gene'], keep='last')
                del db['mm']
                Gdb = pd.merge(Gdb, db, on='gene', how='left')

            db = Gdb
            for c in ['N_sites', 'S_sites']:
                if c in db.columns:
                    del db[c]

            db = reorder_columns(db, column_order)
            db = db[db['coverage'] > 0]

        elif name == 'genome_info':
            column_order = ['genome', 'coverage', 'breadth', 'nucl_diversity',
                            'length', 'true_scaffolds', 'detected_scaffolds',
                            'coverage_median', 'coverage_std', 'coverage_SEM',
                            'breadth_minCov', 'breadth_expected',
                            'nucl_diversity_rarefied',
                            'conANI_reference', 'popANI_reference',
                            'iRep', 'iRep_GC_corrected',
                            'linked_SNV_count', 'SNV_distance_mean', 'r2_mean','d_prime_mean',
                            'consensus_divergent_sites',
                            'population_divergent_sites',
                            'SNS_count', 'SNV_count',
                            'filtered_read_pair_count',
                            'reads_unfiltered_pairs',
                            'reads_mean_PID']

            db = self.get('genome_level_info')
            db = reorder_columns(db, column_order)

            # Get rid of some of those dumb columns
            read_columns = [r for r in db.columns if r.startswith('reads_')]
            keep_columns = [r for r in read_columns if r in ['reads_unfiltered_reads',
                            'reads_unfiltered_pairs', 'reads_mean_PID']]
            for col in set(read_columns) - set(keep_columns):
                if col in db:
                    del db[col]

            if not report_mm_level:
                if 'mm' in db.columns:
                    db = db.sort_values('mm').drop_duplicates(
                            subset=['genome'], keep='last')\
                            .sort_values('genome')
                    del db['mm']

        elif name == 'mapping_info':
            column_order = ['scaffold', 'pass_pairing_filter', 'filtered_pairs']

            db = self.get('mapping_info')
            values = inStrain.filter_reads.write_mapping_info(db, None, **kwargs)

            if store:
                base = self.get_output_base()
                location = base + name + '.tsv'
                os.remove(location) if os.path.exists(location) else None
                f = open(location, 'a')
                f.write("# {0}\n".format(' '.join(["{0}:{1}".format(k, v) for k, v in values.items()])))

                db = reorder_columns(db, column_order)
                db.to_csv(f, index=False, sep='\t')

                f.close()

            if return_table:
                return db
            else:
                return

        elif name == 'comparisonsTable':
            db = self.get_nonredundant_RC_table()

        elif name == 'pairwise_SNP_locations':
            column_order = ['mm', 'scaffold', 'position',
            'name1', 'name2',
            'consensus_SNP', 'population_SNP',
            'con_base_1', 'ref_base_1', 'var_base_1', 'position_coverage_1',
            'A_1', 'C_1', 'T_1', 'G_1',
            'con_base_2', 'ref_base_2', 'var_base_2', 'position_coverage_2',
            'A_2', 'C_2', 'T_2', 'G_2']

            db = self.get('pairwise_SNP_locations')
            db = reorder_columns(db, column_order)

            if ((not report_mm_level) & (len(db) > 0)):
                db = db.sort_values('mm')\
                        .drop_duplicates(subset=['scaffold', 'position',
                                        'name1', 'name2'], keep='last')\
                        .sort_index().drop(columns=['mm'])

        else:
            logging.error("Do not know how to store {0}! Crashing now".format(name))
            assert False

        if store:
            out_base = self.get_output_base()
            db.to_csv(out_base + name + '.tsv', index=False, sep='\t')

        if return_table:
            return db

    def get_parsed_log(self, most_recent=True):
        logloc = os.path.join(self.get_location('log'), 'log.log')
        Ldb = inStrain.logUtils.load_log(logloc)

        # Filter the log
        if most_recent:
            Ldb = inStrain.logUtils.filter_most_recent(Ldb)

        return Ldb

    def get_output_base(self):
        return self.get_location('output') + \
                    os.path.basename(self.get('location')) + '_'

    def get_read_length(self):
        Rdb = self.get('mapping_info').head(1)
        return float(Rdb.loc[0, 'mean_pair_length'])

    def verify(self):
        '''
        Run a series of tests to make sure everything is kosher

        1) No duplicates in the attributes
        2) Everything in the attributes exists
        '''
        pass

    def get_nonredundant_RC_table(self):
        '''
        Get an RC table with just one line per scaffold
        '''
        scdb = self.get('comparisonsTable')
        if (scdb is None) or (len(scdb) == 0):
            return pd.DataFrame()
        else:
            return scdb.sort_values('mm')\
                    .drop_duplicates(subset=['scaffold', 'name1', 'name2'], keep='last')\
                    .sort_index().drop(columns=['mm'])

    def get_nonredundant_scaffold_table(self):
        '''
        Get a scaffold table with just one line per scaffold
        '''
        scdb = self.get('cumulative_scaffold_table')
        if (scdb is None) or (len(scdb) == 0):
            return pd.DataFrame()
        else:
            return scdb.sort_values('mm')\
                    .drop_duplicates(subset=['scaffold'], keep='last')\
                    .sort_index().drop(columns=['mm'])

    def get_nonredundant_snv_table(self):
        '''
        Get a SNP table with just one line per scaffold
        '''
        scdb = self.get('cumulative_snv_table')
        if (scdb is None) or (len(scdb) == 0):
            return pd.DataFrame()
        else:
            return scdb.sort_values('mm')\
                    .drop_duplicates(subset=['scaffold', 'position'], keep='last')\
                    .sort_index().drop(columns=['mm'])

    def get_nonredundant_linkage_table(self):
        '''
        Get a SNP table with just one line per scaffold
        '''
        scdb = self.get('raw_linkage_table')
        if (scdb is None) or (len(scdb) == 0):
            return pd.DataFrame()
        else:
            return scdb.sort_values('mm')\
                    .drop_duplicates(subset=['scaffold', 'position_A', 'position_B'], keep='last')\
                    .sort_index().drop(columns=['mm'])

    def get_clonality_table(self, nonredundant=True):
        '''
        Get a clonality table
        '''
        clonT = self.get('clonT')
        if clonT is None:
            return pd.DataFrame()

        else:
            dbs = []
            scaff2clonT = clonT
            for scaff, clonT in scaff2clonT.items():
                db = inStrain.profile.profile_utilities._clonT_to_table(clonT)
                db['scaffold'] = scaff
                dbs.append(db)

            # The dropna is necessary because clonT doesn't have the "run_up_NaN"
            Cdb = pd.concat(dbs).dropna().reset_index(drop=True)

            if nonredundant:
                Cdb = Cdb.sort_values('mm').dropna()\
                        .drop_duplicates(subset=['scaffold', 'position'], keep='last')\
                        .sort_index().drop(columns=['mm'])

            return Cdb

    def __str__(self):
        '''
        String representation of attributes file
        '''
        string = str(self._get_attributes_file())
        return string

    def _make_fileStructure(self):
        '''
        Make the top level file structure
        '''
        location = self.location

        if not os.path.exists(location):
            os.makedirs(location)

        for l in SNVprofile.firstLevels:
            loc = location + '/' + l
            if not os.path.exists(loc):
                os.makedirs(loc)

    def _initialize_attributes(self):
        '''
        Make an "attributes.tsv" file, and store the location in
        '''
        # This is a new IS object
        if not os.path.exists(os.path.join(self.location, 'raw_data/attributes.tsv')):

            table = {'value':[], 'type':[], 'description':[]}
            Adb = pd.DataFrame(table)
            self._store_attributes_file(Adb)

            # Store the location
            self.store('location', self.location, 'value', 'Location of SNVprofile object')

            # Store the version
            self.store('version', __version__, 'value', 'Version of inStrain')

            # Make a README.txt file
            rm_loc = self._get_fileloc('_README.txt')
            with open(rm_loc, 'w') as o:
                o.write("The data in this folder can be easily accessed using the inStrain python API.\nFor information on how this is done, see the inStrain documentaion at https://instrain.readthedocs.io/en/latest/\n")


        # This is an existing IS object
        else:
            Adb = self._get_attributes_file()

            if not same_versions(Adb.loc['version', 'value'], __version__):
                error_message = "Warning! Your inStrain folder is from version {0}, while the installed version is {1}.\nIf you experience weird behavior, this might be why".format(
                        Adb.loc['version', 'value'], __version__)

                # This is needed because otherwise will make a logger in error!
                if logging.getLogger('').handlers:
                    logging.error(error_message)
                else:
                    print(error_message)

            if self.location != self.get('location'):
                self.store('location', self.location, 'value', 'Location of SNVprofile object')

    def _get_attributes_file(self):
        '''
        Load the attributes file
        '''

        Aloc = str(os.path.join(self.location, 'raw_data/attributes.tsv')).strip()
        # This crazy mumbo-jumbo is to prevent a failure when trying to read while its being written
        for i in range(0,100):
            # while True:
            try:
                Adb = pd.read_csv(Aloc, sep='\t', index_col='name')
                break
            except:
                logging.error("Cannot load {0}; try {1}".format(Aloc, i))
                time.sleep(1)

        return Adb

    def _store_attributes_file(self, Adb):
        '''
        Store the attributes files
        '''
        Adb.to_csv(os.path.join(self.location, 'raw_data/attributes.tsv'),
                                    sep='\t', index_label='name')

    def _get_covt_keys(self):
        '''
        A special method for getting the keys to covT
        '''
        Adb = self._get_attributes_file()

        scaffs = set()
        filename = self.get_location('raw_data') + '/' + os.path.basename(Adb.loc['covT', 'value'])
        f = h5py.File(filename, 'r')
        for thing in list(f.keys()):
            scaff, mm = thing.split('::')
            scaffs.add(scaff)
        return scaffs


    # def _store_special(self, obj, name):
    #     '''
    #     store special things
    #     '''
    #     if name in ['covT', 'snpsCounted', 'clonT']:
    #         '''
    #         The idea here is to save a series of numpy arrays.
    #
    #         Each scaffold has two arrays in a row; one ending in ::covs and one
    #         ending in ::mms. The covs is a 2d array of mm and counts, and mms
    #         is just a list of the true mm levels, so that you can translate
    #         the index of ::covs back to the original
    #         '''
    #         fileloc = self._get_fileloc(name) + '.npz'
    #
    #         to_save = {}
    #         for scaff, covt in obj.items():
    #             if (scaff == 'N5_271_010G1_scaffold_0') & (name == 'snpsCounted'):
    #                 print('storing')
    #
    #             covs = []
    #             mms = []
    #             for mm, cov in covt.items():
    #                 mms.append(mm)
    #                 covs.append(cov)
    #
    #             to_save[scaff + '::covs'] = np.array(covs)
    #             to_save[scaff + '::mms'] = np.array(mms)
    #
    #         np.savez_compressed(fileloc, **to_save)
    #
    #     else:
    #         logging.error("I dont know how to store {0}! Ill just pickle it".format(name))
    #         fileloc = self._get_fileloc(name) + '.pickle'
    #         self._store_pickle(obj, fileloc)
    #
    #     return fileloc
    #
    def _load_special_old(self, location, name):
        '''
        store special things; this is for loading deprecated things
        '''
        if name in ['covT', 'snpsCounted', 'clonT', 'clonTR']:
            base = np.load(location)
            files = base.files


            covT = {}
            # iterate in groups of 2
            for covs, mms in [files[i:i+2] for i in range(0, len(files), 2)]:
                scaff = covs[:-6]
                assert covs[:-6] == mms[:-5], [covs[:-6],  mms[:-5]]

                covT[scaff] = {}
                mm_list = base[mms]
                cov_array = base[covs]

                for i, mm in enumerate(mm_list):
                    covT[scaff][mm] = cov_array[i]

            return covT

        else:
            logging.error("I dont know how to load {0}! Ill just try pickle ".format(name))
            return self._load_pickle(location)

    def _store_special(self, obj, name):
        '''
        store special things using hd5
        '''
        if name in ['covT', 'snpsCounted', 'clonT', 'clonTR']:
            '''
            The idea here is to save the series'
            '''
            fileloc = self._get_fileloc(name) + '.hd5'
            f = h5py.File(fileloc, "w")
            for scaff, clon in obj.items():
                for mm, arr in clon.items():
                    #arr.to_hdf(fileloc, key="{0}::{1}".format(scaff, mm), mode='a')
                    dset = f.create_dataset("{0}::{1}".format(scaff, mm),
                            data=np.array([arr.values, arr.index]),
                            compression="gzip")# convert from series to 2d array

        elif name in ['scaff2pair2mm2SNPs', 'scaff2pair2mm2cov']:
            fileloc = self._get_fileloc(name) + '.hd5'
            if name == 'scaff2pair2mm2cov':
                convert = True
            else:
                convert = False
            store_scaff2pair2mm2SNPs(obj, fileloc, convert=convert)


        else:
            logging.error("I dont know how to store {0}! Ill just pickle it".format(name))
            fileloc = self._get_fileloc(name) + '.pickle'
            self._store_pickle(obj, fileloc)

        return fileloc

    def _load_special(self, location, name, **kwargs):
        '''
        store special things

        with the kwarg "scaffolds", only load certain scaffolds
        '''
        scaffolds = kwargs.get('scaffolds', [])
        pairs = kwargs.get('pairs', [])

        if name in ['covT', 'snpsCounted', 'clonT', 'clonTR']:
            if location[-4:] == '.npz':
                return self._load_special_old(location, name)

            scaff2mm = {}
            f = h5py.File(location, 'r')
            for thing in list(f.keys()):
                scaff, mm = thing.split('::')
                if scaffolds != []:
                    if scaff not in scaffolds:
                        continue

                #pd.read_hdf(location, key=thing)
                dset = list(f[thing])
                mm = int(mm)

                if scaff not in scaff2mm:
                    scaff2mm[scaff] = {}

                scaff2mm[scaff][mm] = pd.Series(data = dset[0],
                                                index = np.array(dset[1].astype('int'))) # convert from 2d array to series

            return scaff2mm

        elif name in ['scaff2pair2mm2SNPs', 'scaff2pair2mm2cov']:
            return load_scaff2pair2mm2SNPs(location, scaffolds=scaffolds, pairs=pairs)

        else:
            logging.error("I dont know how to load {0}! Ill just try pickle ".format(name))
            return self._load_pickle(location)


    def _store_list(self, obj, location):
        '''
        Store a list using txt
        '''
        assert type(obj) == type(list())

        with open(location, "w") as f:
            for s in obj:
                f.write(str(s) +"\n")

    def _load_list(self, location):
        '''
        Load a list using txt
        '''
        l  = []
        with open(location, "r") as f:
            for line in f:
                l.append(line.strip())
        return l

    def _store_numpy(self, obj, location):
        np.savez_compressed(location, obj)

    def _load_numpy(self, location):
        loaded = np.load(location, allow_pickle=True)['arr_0']
        return loaded

    def _store_pandas(self, obj, location):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            obj.to_csv(location)

    def _load_pandas(self, location):
        # This stupid thing is required because just calling the read_csv sometimes causes the warning
        # /home/mattolm/.pyenv/versions/3.8.2/envs/3.8.2_testInstrain/lib/python3.8/site-packages/numpy/lib/
        # arraysetops.py:569: FutureWarning: elementwise comparison failed; returning scalar
        # instead, but in the future will perform elementwise comparison
        # mask |= (ar1 == a)"
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return pd.read_csv(location, index_col=0)

    def _store_pickle(self, obj, location):
        f = open(location, 'wb')
        pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)
        f.close()

    def _load_pickle(self, location):
        f = open(location, 'rb')
        tmp_dict = pickle.load(f)
        f.close()
        return tmp_dict

    def _store_dictionary(self, dic, location):
        '''
        Store a dictionary using json
        '''
        assert type(dic) == type({})

        with open(location, 'w') as fp:
            json.dump(dic, fp, default=_json_helper)

    def _load_dictionary(self, location):
        '''
        Load a dictionary using json
        '''
        with open(location, 'r') as fp:
            return(json.load(fp))

    def _get_fileloc(self, name):
        return os.path.join(self.location, 'raw_data/{0}'.format(name))

def same_versions(v1, v2):
    '''
    If the major or minor version are the same, return True; else, False
    '''
    minor1 = int(v1.split('.')[1])
    minor2 = int(v2.split('.')[1])

    major1 = int(v1.split('.')[0])
    major2 = int(v2.split('.')[0])

    return (minor1 == minor2) & (major1 == major2)

def store_scaff2pair2mm2SNPs(obj, fileloc, convert=False):
    '''
    Store as an hdf5 object

    If convert, declare it as a numpy array from a set
    '''
    f = h5py.File(fileloc, "w")
    for scaff, pair2mm2SNPs in obj.items():
        for pair, mm2SNPs in pair2mm2SNPs.items():
            for mm, arr in mm2SNPs.items():
                if convert:
                    dset = f.create_dataset("{0}::{1}::{2}".format(scaff, pair, mm),
                            data=np.fromiter(arr, int, len(arr)), compression="gzip")
                else:
                    dset = f.create_dataset("{0}::{1}::{2}".format(scaff, pair, mm),
                            data=np.array(arr),compression="gzip")

def load_scaff2pair2mm2SNPs(location, scaffolds=[], pairs=[]):
    scaff2pair2mm2SNPs = {}
    f = h5py.File(location, 'r')

    for thing in list(f.keys()):
        scaff, pair, mm = thing.split('::')

        if scaffolds != []:
            if scaff not in scaffolds:
                continue

        if pairs != []:
            if pair not in pairs:
                continue

        dset = f[thing].value
        mm = int(mm)

        if scaff not in scaff2pair2mm2SNPs:
            scaff2pair2mm2SNPs[scaff] = {}

        if pair not in scaff2pair2mm2SNPs[scaff]:
            scaff2pair2mm2SNPs[scaff][pair] = {}

        scaff2pair2mm2SNPs[scaff][pair][mm] = dset # convert from 2d array to series

    return scaff2pair2mm2SNPs

def _json_helper(o):
    if isinstance(o, np.int64): return int(o)
    if isinstance(o, Bio.Seq.Seq): return str(o)
    raise TypeError

class SNVprofile_old:
    '''
    The class holds the profile of a single .fasta / .bam pair
    '''

    def __init__(self, **kwargs):
        '''
        initialize all attributes to None
        '''

        self.ATTRIBUTES = (\
        'filename', # Filename of this object

        'fasta_loc',
        'scaffold2length', # Dictionary of scaffold 2 length
        'bam_loc',

        'scaffold_list', # 1d list of scaffolds, in same order as counts_table
        'counts_table', #1d numpy array of 2D counts tables for each scaffold
        'raw_snp_table', # Contains raw SNP information on a mm level
        'raw_ANI_table', # Contains raw ANI information on a mm level
        'raw_coverage_table', # Contains raw coverage information on a mm level
        'raw_linkage_table', # Contains raw linkage information on a mm level

        'cumulative_scaffold_table', # Cumulative coverage on mm level. Formerly "scaffoldTable.csv"
        'cumulative_snv_table', # Cumulative SNP on mm level. Formerly "snpLocations.pickle"

        # THE FOLLOWING ARE HEAVY-RAM OBJECTS
        'scaff2covT',
        'scaff2basesCounted',
        'scaff2snpsCounted',
        )
        for att in self.ATTRIBUTES:
            setattr(self, att, kwargs.get(att, None))
        self.version = __version__

    def __str__(self):
        string = '\n'.join(["{0} - {1}".format(att, type(getattr(self, att)))\
            for att in self.ATTRIBUTES])
        return string

    def get_nonredundant_scaffold_table(self):
        '''
        Get a scaffold table with just one line per scaffold
        '''
        if (self.cumulative_scaffold_table is None) or (len(self.cumulative_scaffold_table) == 0):
            return pd.DataFrame()
        else:
            return self.cumulative_scaffold_table.sort_values('mm')\
                    .drop_duplicates(subset=['scaffold'], keep='last')\
                    .sort_index().drop(columns=['mm'])

    def get_nonredundant_linkage_table(self):
        '''
        Get a SNP table with just one line per scaffold
        '''
        if (self.raw_linkage_table is None) or (len(self.raw_linkage_table) == 0):
            return pd.DataFrame()
        else:
            return self.raw_linkage_table.sort_values('mm')\
                    .drop_duplicates(subset=['scaffold', 'position_A', 'position_B'], keep='last')\
                    .sort_index().drop(columns=['mm'])

    def get_nonredundant_snv_table(self):
        '''
        Get a SNP table with just one line per scaffold
        '''
        if (self.cumulative_snv_table is None) or (len(self.cumulative_snv_table) == 0):
            return pd.DataFrame()
        else:
            return self.cumulative_snv_table.sort_values('mm')\
                    .drop_duplicates(subset=['scaffold', 'position'], keep='last')\
                    .sort_index().drop(columns=['mm'])

    def get_clonality_table(self, nonredundant=True):
        '''
        Get a clonality table
        '''
        if not hasattr(self, 'clonT'):
            return pd.DataFrame()

        elif self.clonT is None:
            return pd.DataFrame()

        else:
            dbs = []
            scaff2clonT = self.clonT
            for scaff, clonT in scaff2clonT.items():
                db = inStrain.profile.profile_utilities._clonT_to_table(clonT)
                db['scaffold'] = scaff
                dbs.append(db)

            # The dropna is necessary because clonT doesn't have the "run_up_NaN"
            Cdb = pd.concat(dbs).dropna().reset_index(drop=True)

            if nonredundant:
                Cdb = Cdb.sort_values('mm').dropna()\
                        .drop_duplicates(subset=['scaffold', 'position'], keep='last')\
                        .sort_index().drop(columns=['mm'])

            return Cdb

    def store(self):
        '''
        Store self. MUST have the attribute "filename" set
        '''
        if self.filename is None:
            print("Cant save this SNVprofile- no filename!")
            return
        elif self.filename[-7:] == '.pickle':
            self.filename = self.filename[:-7]

        f = open(self.filename + ".pickle", 'wb')
        pickle.dump(self.__dict__, f, pickle.HIGHEST_PROTOCOL)
        f.close()

    def load(self, basename):
        '''
        Load self from the basename
        '''
        if basename[-7:] == '.pickle':
            basename = basename[:-7]

        #print("loading {0}".format(basename + '.pickle'))
        f = open(basename + '.pickle', 'rb')
        tmp_dict = pickle.load(f)
        f.close()

        self.__dict__.clear()
        self.__dict__.update(tmp_dict)

        return self

def convert_SNVprofile(pickle_loc):
    '''
    From the old SNVprofile type (a pickle), make a new one
    '''
    # Load the old version
    oIS = SNVprofile_old()
    oIS.load(pickle_loc)

    # Get a new version set up
    nIS = SNVprofile(pickle_loc.replace('.pickle', ''))

    for attr in oIS.__dict__.keys():
    #for attr in oIS.ATTRIBUTES:
        #print('storing {0}'.format(attr))
        if attr in ['ATTRIBUTES', 'filename']:
            continue

        elif getattr(oIS, attr) is None:
            pass

        elif attr == 'version':
            nIS.store('old_version', getattr(oIS, attr), 'value', 'Version of old SNVprofile this was made from')

        elif attr == 'fasta_loc':
            # THIS IS INCORRECTLY SAVED AT TH MOMEMENT! - M.O. 6/6/19
            #nIS.store('fasta_loc', oIS.fasta_loc, 'value', 'Location of .fasta file')
            nIS.store('fasta_loc', 'unk', 'value', 'Location of .fasta file')

        elif attr == 'scaffold2length':
            nIS.store('scaffold2length', oIS.scaffold2length, 'dictionary', 'Dictionary of scaffold 2 length')

        elif attr == 'bam_loc':
            nIS.store('bam_loc', getattr(oIS, attr), 'value', 'Location of .bam file')

        elif attr == 'scaffold_list':
            nIS.store('scaffold_list', getattr(oIS, attr), 'list', '1d list of scaffolds, in same order as counts_table')

        elif attr == 'counts_table':
            nIS.store('counts_table', getattr(oIS, attr), 'numpy', '1d numpy array of 2D counts tables for each scaffold')

        elif attr == 'raw_snp_table':
            nIS.store('raw_snp_table', getattr(oIS, attr), 'pandas', 'Contains raw SNP information on a mm level')

        elif attr == 'raw_ANI_table':
            assert getattr(oIS, attr) == None
            #nIS.store('raw_ANI_table', getattr(oIS, attr), 'pandas', 'Contains raw ANI information on a mm level')

        elif attr == 'raw_coverage_table':
            assert getattr(oIS, attr) == None

        elif attr == 'raw_linkage_table':
            nIS.store('raw_linkage_table', getattr(oIS, attr), 'pandas', 'Contains raw linkage information on a mm level')

        elif attr == 'cumulative_scaffold_table':
            nIS.store('cumulative_scaffold_table', getattr(oIS, attr), 'pandas', "Cumulative coverage on mm level. Formerly scaffoldTable.csv")

        elif attr == 'cumulative_snv_table':
            nIS.store('cumulative_snv_table', getattr(oIS, attr), 'pandas', "Cumulative SNP on mm level. Formerly snpLocations.pickle")
        #
        # '''
        # THOSE DON'T WORK BECAUSE THE BASEWISE SHRINK ISN'T MEANT TO WORK WHEN YOU HAVE A SCAFFOLD FISRT!!
        # '''
        #

        elif attr == 'covT':
            new = {s:inStrain.profile.profile_utilities.shrink_basewise(cov, 'coverage') for s, cov in getattr(oIS, attr).items()}
            nIS.store('covT', new,
                    'special', "Scaffold -> mm -> position based coverage")

        elif attr == 'snpsCounted':
            new = {s:inStrain.profile.profile_utilities.shrink_basewise(cov, 'snpCounted') for s, cov in getattr(oIS, attr).items()}
            nIS.store('snpsCounted', new,
                    'special', "Scaffold -> mm -> position based True/False on if a SNPs is there")

        elif attr == 'clonT':
            new = {s:inStrain.profile.profile_utilities.shrink_basewise(cov, 'clonality') for s, cov in getattr(oIS, attr).items()}
            nIS.store('clonT', new,
                    'special', "Scaffold -> mm -> position based clonality")

        elif attr == 'clonTR':
            new = {s:inStrain.profile.profile_utilities.shrink_basewise(cov, 'clonality') for s, cov in getattr(oIS, attr).items()}
            nIS.store('clonTR', new,
                    'special', "Scaffold -> mm -> position based clonality")

        elif attr in ['mapping_info', 'read_report']:
            nIS.store(attr, getattr(oIS, attr), 'pandas', "Report on reads")

        else:
            logging.error('I dont know how to store {0}!'.format(attr))
            print(type(getattr(oIS, attr)))
            break

def reorder_columns(db, column_order):
    '''
    Reorder columns in db baesd on column order

    Any column not in the column_order will be added to the end
    '''
    columns = set(db.columns)
    return db[[c for c in column_order if c in columns] \
                + list(columns - set(column_order))]

# if __name__ == '__main__':
#     parser = argparse.ArgumentParser(description= """
#         A quick way to convert old SNP profiles (v0.3.x) to new SNP profiles (v0.4.x)""")
#
#     # Required positional arguments
#     parser.add_argument('input', help="an on inStrain object pickle")
#     args = parser.parse_args()
#
#     from _version import __version__
#
#     convert_SNVprofile(args.input)



#!/usr/bin/env python

import os
import csv
import sys
import glob
import scipy
import logging
import warnings
import argparse
import traceback

import numpy as np
import pandas as pd
from tqdm import tqdm
from Bio import SeqIO
import concurrent.futures
from concurrent import futures
from inStrain import SNVprofile
from collections import defaultdict

import inStrain.SNVprofile
import inStrain.controller
import inStrain.profile.profile_utilities
import inStrain.quickProfile
import inStrain.irep_utilities
import inStrain.logUtils

class Controller():

    def main(self, args):
        '''
        The main method when run on the command line
        '''
        # Parse arguments
        IS, vargs = self.validate_input(args)

        # Load scaffold to bin and bin to length; put in IS
        logging.debug('Loading scaffold to bin')
        self.prepare_genome_wide(IS, vargs)

        # Make the genome-wide information table for a profile object
        object_type = IS.get('object_type')
        if object_type is None:
            if IS.get('comparisonsTable') is None:
                object_type = 'profile'
            else:
                object_type = 'compare'
            # logging.error("Theres no object type! This is bad! I'll assume profile")
            # print(IS)
            # object_type = 'profile'

        if object_type == 'profile':
            # Generate and store the genome level info
            GIdb = genomeLevel_from_IS(IS, **vargs)
            IS.store('genome_level_info', GIdb, 'pandas', 'Table of genome-level information')

            # Store the output
            IS.generate('genome_info', **vargs)
            IS.generate('SNVs')

        else:
            self.prepare_genome_wide(IS, vargs)
            gdb = genomeWideFromIS(IS, 'read_compaer', **vargs)
            out_base = IS.get_location('output') + os.path.basename(IS.get('location')) + '_'
            gdb.to_csv(out_base + 'genomeWide_compare.tsv', index=False, sep='\t')
            # out_base = IS.get_location('output') + os.path.basename(IS.get('location')) + '_'
            # GIdb.to_csv(out_base + 'genome_info.tsv', index=False, sep='\t')

            # Store the SNV output
            # IS.generate('SNVs')

        # Make the genome-wide information table for a compare object
        # elif object_type == 'compare':
        #     pass
        #
        # # Figure out output base
        # out_base = IS.get_location('output') + os.path.basename(IS.get('location')) + '_'
        #
        #
        # # Do genome info
        # try:
        #     gdb = genomeWideFromIS(IS, 'scaffold_info', mm_level=mm_level)
        #     gdb.to_csv(out_base + 'genomeWide_scaffold_info.tsv', index=False, sep='\t')
        # except:
        #     logging.debug("GenomeWide scaffold info failed ", exc_info=1)
        #     #traceback.print_exc()
        #
        # # Do read report
        # try:
        #     gdb = genomeWideFromIS(IS, 'mapping_info', mm_level=mm_level)
        #     gdb.to_csv(out_base + 'genomeWide_mapping_info.tsv', index=False, sep='\t')
        # except:
        #     logging.debug("GenomeWide read report failed ", exc_info=1)
        #     #traceback.print_exc()
        #
        # # Do read comparer
        # try:
        #     gdb = genomeWideFromIS(IS, 'read_compaer', mm_level=mm_level)
        #     gdb.to_csv(out_base + 'genomeWide_compare.tsv', index=False, sep='\t')
        # except:
        #     logging.debug("GenomeWide compare failed ", exc_info=1)
        #     # traceback.print_exc()

    def validate_input(self, args):
        '''
        Validate and mess with the arguments a bit
        '''
        # Make sure the IS object is OK
        assert os.path.exists(args.IS)
        IS = inStrain.SNVprofile.SNVprofile(args.IS)

        # Set up the logger
        log_loc = IS.get_location('log') + 'log.log'
        inStrain.controller.setup_logger(log_loc)

        # Convert argument type
        vargs = vars(args)
        vargs.pop('IS')

        return IS, vargs

    def prepare_genome_wide(self, IS, vargs):
        '''
        Load and store the scaffold to bin file and bin to length file
        '''
        stb = load_scaff2bin(vargs.get('stb'), IS)

        # Make bin to length
        stl = IS.get('scaffold2length')
        b2l = {}
        for scaffold, bin in stb.items():
            if bin not in b2l:
                b2l[bin] = 0

            if scaffold in stl:
                b2l[bin] += stl[scaffold]
            else:
                logging.error("FAILURE StbError {0} {1} no_length will not be considered as part of the genome".format(scaffold, bin))

        # Store these
        IS.store('scaffold2bin', stb, 'dictionary', 'Dictionary of scaffold 2 bin')
        IS.store('bin2length', b2l, 'dictionary', 'Dictionary of bin 2 total length')

def genomeLevel_from_IS(IS, **kwargs):
    '''
    Calculate genome-wide metrics based on scaffold-level info.

    IS object must already have:
        scaffold2bin
        bin2length

    Arguments:
        IS = InStrain Profile object; must already be initialized

    kwargs:
        mm-level = If True, calculate all metrics on the mm-level
    '''
    inStrain.logUtils.log_checkpoint("GenomeLevel", "genomeLevel_from_IS", "start")

    # Load necessary stuff from IS file
    stb = IS.get('scaffold2bin')
    b2l = IS.get('bin2length')
    s2l = IS.get('scaffold2length')

    # Calculate averaing and summing metrics from the scaffold table
    db = IS.get('cumulative_scaffold_table')
    gdb = _add_stb(db, stb)

    # Handle mm level
    skip_mm_level = kwargs.get('skip_mm_profiling', False)
    if skip_mm_level:
        gdb = gdb.sort_values('mm').drop_duplicates(
                subset=['scaffold'], keep='last')\
                .sort_values('scaffold')
        gdb['mm'] = 1000

    # Calculate genome level scaffold info
    inStrain.logUtils.log_checkpoint("GenomeLevel", "scaffold_info", "start")

    GSI_db = _genomeLevel_scaffold_info_v3(gdb, stb, b2l, **kwargs)

    inStrain.logUtils.log_checkpoint("GenomeLevel", "scaffold_info", "end")

    # Prepare to calculate coverage distribution metrics
    table = defaultdict(list)
    bin2scaffolds = calc_bin2scaffols(stb)
    scaff2sequence = _load_scaff2sequence(IS, **kwargs)

    relevant_genomes = calc_relevant_genomes(GSI_db, IS, **kwargs)
    relevant_scaffolds = set()
    for g in relevant_genomes:
        for s in bin2scaffolds[g]:
            relevant_scaffolds.add(s)
    covT = IS.get('covT', scaffolds=relevant_scaffolds)

    # Figure out total number of mms
    if skip_mm_level:
        mms = [1000]
    else:
        mms = calc_mms(covT)

    # Calculate expensive coverage distribution metrics
    inStrain.logUtils.log_checkpoint("GenomeLevel", "coverage_info", "start")

    EG_db = genomeLevel_coverage_info(covT, bin2scaffolds, relevant_genomes,
                                        s2l, scaff2sequence, mms, **kwargs)

    inStrain.logUtils.log_checkpoint("GenomeLevel", "coverage_info", "end")

    # Calculate genome-level read mapping
    rdb = IS.get('mapping_info')
    rdb = rdb[rdb['scaffold'] != 'all_scaffolds']
    rdb = _add_stb(rdb, stb)

    inStrain.logUtils.log_checkpoint("GenomeLevel", "mapping_info", "start")

    rdb = _genome_wide_rr(rdb, stb, **kwargs)
    rdb = rdb.rename(columns={'reads_filtered_pairs':'filtered_read_pair_count'})
    if 'reads_pass_pairing_filter' in rdb.columns:
        del rdb['reads_pass_pairing_filter']

    inStrain.logUtils.log_checkpoint("GenomeLevel", "mapping_info", "end")

    # Merge
    mdb = pd.merge(GSI_db, EG_db, on=['genome', 'mm'], how='outer')
    mdb = pd.merge(mdb, rdb, on=['genome'], how='left')

    # Calculate genome-level linkage metrics
    ldb = IS.get('raw_linkage_table')
    if len(ldb) > 0:
        if skip_mm_level:
            ldb = ldb.sort_values('mm').drop_duplicates(
                    subset=['scaffold', 'position_A', 'position_B'], keep='last')\
                    .sort_values(['scaffold', 'position_A', 'position_B'])
            gdb['mm'] = 1000

        ldb = _add_stb(ldb, stb)
        if (ldb is not None) and (len(ldb) > 0):

            inStrain.logUtils.log_checkpoint("GenomeLevel", "linkage", "start")

            ldb = _genome_wide_linkage(ldb, stb, mms, **kwargs)

            inStrain.logUtils.log_checkpoint("GenomeLevel", "linkage", "end")


            mdb = pd.merge(mdb, ldb, on=['genome', 'mm'], how='left')


    if skip_mm_level:
        del mdb['mm']

    inStrain.logUtils.log_checkpoint("GenomeLevel", "genomeLevel_from_IS", "end")

    return mdb

def _load_scaff2sequence(IS, **kwargs):
    '''
    Try and get scaff2sequence to GC-correct iRep. Return None if you cant
    '''
    if kwargs.get('fasta', None) is not None:
        fasta_loc = kwargs.get('fasta')
    else:
        fasta_loc = IS.get('fasta_loc')

    if fasta_loc is None:
        logging.error("Do not have a .fasta file; will not GC correct iRep")
        scaff2sequence = None
    else:
        try:
            scaff2sequence = SeqIO.to_dict(SeqIO.parse(fasta_loc, "fasta"))
        except:
            logging.error("Could not load .fasta file {0}; will not GC correct iRep".format(fasta_loc))
            scaff2sequence = None
    return scaff2sequence

def calc_relevant_genomes(GSI_db, IS, **kwargs):
    '''
    Depending on kwrags, determine which scaffolds should have expensive profiling done
    '''
    return set(GSI_db['genome'].tolist())

def genomeLevel_coverage_info(covT, bin2scaffolds, relevant_genomes, s2l,
                                scaff2sequence, mms, **kwargs):
    '''
    THIS IS ALL CALCULATED WITH MASKED EDGES OF SCAFFOLDS
    '''

    dbs = []

    for genome, scaffolds in bin2scaffolds.items():
        if genome not in relevant_genomes:
            continue

        table = defaultdict(list)

        # Sort the scaffolds
        scaffolds = scaffolds.intersection(set(s2l.keys())) # THIS SHOULDN'T HAVE TO BE HERE!
        scaffolds = sorted(scaffolds, key=s2l.get, reverse=True)

        # Try and get the GC_correction
        gc_windows = None
        if scaff2sequence is not None:
            try:
                gc_windows = inStrain.irep_utilities.generate_gc_windows(scaffolds, scaff2sequence, mask_edges=100)
            except:
                pass

        # Set up default iRep
        iRep = np.nan
        iRep_accessory = {'iRep_GC_corrected':np.nan}

        # Iterate for this genome
        for mm in mms:

            # Make a coverage array for this genome in it's entirety
            covs, scaff2genome_index = generate_genome_coverage_array(covT, s2l, order=scaffolds, maxMM=mm, mask_edges=100)

            # Calculate iRep only on mm == 1, or if not on mm level
            if ((mm == 1) or (mm == 1000)):
                try:
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore")
                        iRep, iRep_accessory = inStrain.irep_utilities.calculate_iRep_from_coverage_array(covs, len(scaffolds), gc_windows)
                    logging.debug("iRep {0} {1} {2}".format(genome, mm, iRep_accessory))
                except:
                    logging.warning("FAILURE iRepError {0} {1}".format(genome, mm))
                    iRep = np.nan
                    iRep_accessory = {'iRep_GC_corrected':np.nan}

            # Calculate medians and other variants
            if len(covs) == 0:
                covs = pd.Series([0])

            table['mm'].append(mm)
            table['genome'].append(genome)

            # Dont print annoying runtime warnings here
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                table['coverage_median'].append(int(np.median(covs)))
                table['coverage_SEM'].append(scipy.stats.sem(covs))
                table['coverage_std'].append(np.std(covs))

        gdb = pd.DataFrame(table)
        gdb['iRep'] = iRep
        gdb.loc[:,'iRep_GC_corrected'] = iRep_accessory['iRep_GC_corrected']
        dbs.append(gdb)

    adb = pd.concat(dbs).reset_index(drop=True)
    return adb

def genomeWideFromIS(IS, thing, **kwargs):
    '''
    This is used by plotting utilities
    '''
    stb = IS.get('scaffold2bin')
    b2l = IS.get('bin2length')
    mm_level = kwargs.get('mm_level', False)

    if stb == None:
        logging.error("This IS object does not have an .stb file; cant use it")
        assert False

    if thing == 'scaffold_info':
        db = IS.get('cumulative_scaffold_table')
        gdb = _add_stb(db, stb)
        assert _validate_b2l(gdb, b2l)
        return _genome_wide_si_2(gdb, stb, b2l, **kwargs)

    if thing == 'mapping_info':
        db = IS.get('mapping_info')
        db = db[db['scaffold'] != 'all_scaffolds']
        gdb = _add_stb(db, stb)
        return _genome_wide_rr(gdb, stb, **kwargs)

    if thing == 'read_compaer':
        db = IS.get('comparisonsTable')
        gdb = _add_stb(db, stb)
        return _genome_wide_readComparer(gdb, stb, b2l, **kwargs)

    else:
        logging.error('{0} is not possible to get from genomeWideIS'.format(thing))
        assert False

def makeGenomeWide(thing, db, stb, b2l=None, **kwargs):
    '''
    Convert scaffold-level things to be genome-level things
    '''
    assert type(stb) == type({})

    # # Load it if you have to
    # if thing == 'mm_genome_info':
    #     db = _get_mm_db(db)

    # Validate the stb
    gdb = _add_stb(db, stb)

    # Validate the b2l
    if b2l != None:
        if not _validate_b2l(gdb, b2l):
            return

#     if thing == 'mm_genome_info':
#         return _genome_wide_mm_genomeInfo(gdb, stb, b2l, **kwargs)

    if thing == 'mapping_info':
        return _genome_wide_rr(gdb, stb, **kwargs)

    elif thing == 'scaffold_info':
        return _genome_wide_si_2(gdb, stb, b2l, **kwargs)

    elif thing == 'readComparer':
        return _genome_wide_readComparer(gdb, stb, b2l, **kwargs)

def _add_stb(db, stb, verbose=True):
    gdb = db.copy()

    if len(gdb) == 0:
        logging.error('Error- no scaffolds detected.')
        return

    gdb.loc[:,'genome'] = gdb['scaffold'].map(stb)

    if len(gdb['genome'].dropna().unique()) == 0:
        logging.error('Error- no genomes detected. Example: stb has scaffold {0}, database has scaffold {1}'.format(
                list(stb.keys())[0], db['scaffold'].tolist()[0]))
        return

    if len(gdb[gdb['genome'].isna()]) > 0:
        if verbose:
            logging.info("{0:.2f}% of scaffolds have a genome".format((len(gdb[~gdb['genome'].isna()])/\
                                                          len(gdb))*100))
    return gdb

def _validate_b2l(gdb, b2l):
    genomes = list(gdb['genome'].unique())
    missing = [g for g in genomes if g not in b2l]

    if len(missing) > 1:
        logging.error("{0} of {1} genomes are missing from b2l".format(len(missing), len(genomes)))
        return False

    return True

def _genome_wide_si_2(gdb, stb, b2l, **kwargs):
    '''
    This version can handle mm level
    '''
    skip_mm_level = kwargs.get('skip_mm_level', False)

    if skip_mm_level:
        gdb = gdb.sort_values('mm').drop_duplicates(
                subset=['scaffold'], keep='last')\
                .sort_values('scaffold')
        gdb['mm'] = 0

    table = defaultdict(list)
    for mm in sorted(list(gdb['mm'].unique())):
        Odb = gdb[gdb['mm'] <= mm].sort_values('mm').drop_duplicates(subset=['scaffold'], keep='last')
        for genome, df in Odb.groupby('genome'):
            cols = list(df.columns)

            table['mm'].append(mm)
            table['genome'].append(genome)

            # Scaffolds
            table['detected_scaffolds'].append(len(df))
            table['true_scaffolds'].append(len([True for s, b in stb.items() if b == genome]))
            table['length'].append(int(b2l[genome]))

            # The summing columns
            for col in ['SNPs', 'Referece_SNPs', 'BiAllelic_SNPs', 'MultiAllelic_SNPs', 'consensus_SNPs', 'population_SNPs']:
                if col in cols:
                    table[col].append(df[col].fillna(0).sum())

            # Weighted average (over total length)
            for col in ['breadth', 'coverage', 'coverage_std']:
                table[col].append(sum([x * y for x, y in zip(df[col].fillna(0), df['length'])]) / b2l[genome])

            # Weighted average (over detected scaffold length)
            df.loc[:,'considered_length'] = [x*y for x,y in zip(df['breadth_minCov'], df['length'])]
            considered_leng = float(df['considered_length'].sum())

            # To maintain backwards compatibility
            for col in ['mean_clonality']:
                if col not in df.columns:
                    continue
                if considered_leng != 0:
                    table[col].append(sum(x * y for x, y in zip(df[col].fillna(0), df['considered_length'])) / considered_leng)
                else:
                    table[col].append(np.nan)

            for col in ['mean_microdiverstiy', 'nucl_diversity_rarefied']:
                if col not in df.columns:
                    continue
                if considered_leng != 0:
                    table[col].append(sum(x * y for x, y in zip(df[col].fillna(0), df['considered_length'])) / considered_leng)
                else:
                    table[col].append(np.nan)

            # ANI
            if 'consensus_SNPs' in cols:
                if considered_leng != 0:
                    table['conANI_reference'].append((considered_leng - df['consensus_SNPs'].sum()) / considered_leng)
                    table['popANI_reference'].append((considered_leng - df['population_SNPs'].sum()) / considered_leng)
                else:
                    table['conANI_reference'].append(0)
                    table['popANI_reference'].append(0)
            else:
                if considered_leng != 0:
                    table['ANI'].append((considered_leng - df['SNPs'].sum()) / considered_leng)
                else:
                    table['ANI'].append(0)
            #table['detected_scaff_length'].append(df['length'].sum())

            table['breadth_minCov'].append(considered_leng / b2l[genome])
            table['breadth_expected'].append(estimate_breadth(table['coverage'][-1]))

    db = pd.DataFrame(table)

    # Add back microdiversity
    if (('nucl_diversity' not in df.columns) & ('mean_clonality' in df.columns)):
        db['nucl_diversity'] = 1 - db['mean_clonality']

    if skip_mm_level:
        del db['mm']

    return db

def _genomeLevel_scaffold_info_v3(gdb, s2b, b2l, **kwargs):
    '''
    Private method for calculating genome-level scaffold info
    '''
    stb = s2b # This is because you also have the kwarg
    table = defaultdict(list)
    for mm in sorted(list(gdb['mm'].unique())):
        # Get scaffold information at this mm
        Odb = gdb[gdb['mm'] <= mm].sort_values('mm')\
                .drop_duplicates(subset=['scaffold'], keep='last')

        for genome, df in Odb.groupby('genome'):
            cols = list(df.columns)

            table['mm'].append(mm)
            table['genome'].append(genome)

            # Scaffolds
            table['detected_scaffolds'].append(len(df))
            table['true_scaffolds'].append(len([True for s, b in stb.items()
                                                                if b == genome]))
            table['length'].append(int(b2l[genome]))

            # The summing columns
            for col in ['SNS_count', 'SNV_count', 'divergent_site_count',
                    'consensus_divergent_sites', 'population_divergent_sites']:
                if col in cols:
                    table[col].append(df[col].fillna(0).sum())

            # Weighted average (over total length)
            for col in ['breadth', 'coverage']:
                table[col].append(sum([x * y for x, y in zip(df[col].fillna(0), df['length'])]) / b2l[genome])

            # Weighted average (over detected scaffold length)
            df['considered_length'] = [x*y for x,y in zip(
                                        df['breadth_minCov'], df['length'])]
            considered_leng = float(df['considered_length'].sum())

            for col in ['nucl_diversity', 'nucl_diversity_rarefied']:
                if col not in df.columns:
                    continue
                if considered_leng != 0:
                    table[col].append(sum(x * y for x, y in zip(df[col].fillna(0), df['considered_length'])) / considered_leng)
                else:
                    table[col].append(np.nan)

            # ANI
            if 'consensus_divergent_sites' in cols:
                if considered_leng != 0:
                    table['conANI_reference'].append((considered_leng - df['consensus_divergent_sites'].sum()) / considered_leng)
                    table['popANI_reference'].append((considered_leng - df['population_divergent_sites'].sum()) / considered_leng)
                else:
                    table['conANI_reference'].append(0)
                    table['popANI_reference'].append(0)

            # Special
            table['breadth_minCov'].append(considered_leng / b2l[genome])
            table['breadth_expected'].append(estimate_breadth(table['coverage'][-1]))

    db = pd.DataFrame(table)
    return db

# def _get_mm_db(obj):
#     '''
#     From an inStrain object (currently a pickle), get the raw scaffold table
#     '''
#     s = inStrain.SNVprofile()
#     s.load(obj)
#     db = s.cumulative_scaffold_table
#
#     return db

def _genome_wide_rr(gdb, s2b, **kwrags):
    '''
    THIS REALLY SHOULD USE SCAFFOLD LENGTH FOR A WEIGHTED AVERAGE!
    '''
    stb = s2b
    table = defaultdict(list)
    for genome, df in gdb.groupby('genome'):
        table['genome'].append(genome)
        #table['scaffolds'].append(len(df))
        for col in [c for c in list(df.columns) if c not in ['scaffold', 'genome']]:
            if len(df[col].dropna()) == 0:
                table['reads_' + col].append(np.nan)
            elif col.startswith('pass') or col.startswith('unfiltered_') or col.startswith('filtered'):
                table['reads_' + col].append(df[col].sum())
            else:
                table['reads_' + col].append(df[col].mean())

    return pd.DataFrame(table)

def _genome_wide_linkage(ldb, s2b, mms, **kwrags):
    '''
    From the raw_linkage_table, calculate average d' and r2 for each genome
    '''
    stb = s2b
    table = defaultdict(list)
    for mm in mms:
        # Get scaffold information at this mm
        Odb = ldb[ldb['mm'] <= mm].sort_values('mm')\
                .drop_duplicates(subset=['scaffold', 'position_A', 'position_B'],
                keep='last')

        if len(Odb) == 0:
            continue

        for genome, df in Odb.groupby('genome'):
            table['genome'].append(genome)
            table['mm'].append(mm)
            table['r2_mean'].append(df['r2'].mean())
            table['d_prime_mean'].append(df['d_prime'].mean())
            table['SNV_distance_mean'].append(df['distance'].mean())
            table['linked_SNV_count'].append(len(df))

    return pd.DataFrame(table)

# def _genome_wide_si(gdb, stb, b2l, **kwargs):
#     table = defaultdict(list)
#     for genome, df in gdb.groupby('genome'):
#         table['genome'].append(genome)

#         # Scaffolds
#         table['detected_scaffolds'].append(len(df))
#         table['true_scaffolds'].append(len([True for s, b in stb.items() if b == genome]))
#         table['length'].append(int(b2l[genome]))

#         # The summing columns
#         for col in ['SNPs']:
#             table[col].append(df[col].sum())

#         # Weighted average (over total length)
#         for col in ['breadth', 'coverage', 'coverage_std']:
#             table[col].append(sum(x * y for x, y in zip(df[col], df['length'])) / b2l[genome])

#         # Weighted average (over detected scaffold length)
#         df['considered_length'] = [x*y for x,y in zip(df['breadth_minCov'], df['length'])]
#         considered_leng = df['considered_length'].sum()
#         for col in ['mean_clonality']:
#             table[col].append(sum(x * y for x, y in zip(df[col], df['considered_length'])) / considered_leng)

#         # ANI
#         if considered_leng != 0:
#             table['ANI'].append((considered_leng - df['SNPs'].sum()) / considered_leng)
#         else:
#             table['ANI'].append(0)
#         #table['detected_scaff_length'].append(df['length'].sum())

#         table['breadth_minCov'].append(considered_leng / b2l[genome])
#         table['breadth_expected'].append(estimate_breadth(table['coverage'][-1]))

#     return pd.DataFrame(table)

# def _genome_wide_mm_genomeInfo(gdb, stb, b2l, **kwargs):
#     '''
#     JupyterNotebooks/Infant_Eukaryotes/Euks_13_mappingListGamma_bams_2_load_v2_2.ipynb
#     '''
#     gdb['genome'] = gdb['scaffold'].map(stb)
#     gdb['considered_length'] = [x*y for x,y in zip(gdb['breadth_minCov'], gdb['length'])]

#     # Add blanks for scaffolds that aren't there
#     #bdb = pd.DataFrame({s:})

#     table = defaultdict(list)
#     for genome, mm, db in interate_sdb_mm(gdb, s2l=None, stb=stb):
#         table['genome'].append(genome)
#         table['mm'].append(mm)

#         # Sum columns
#         for col in ['bases_w_0_coverage', 'length']:
#             table[col].append(db[col].sum())

#         # Max columns
#         for col in ['SNPs']:
#             table[col].append(db[col].max())

#         # Weighted average columns
#         for col in ['breadth', 'coverage', 'coverage_std', 'breadth_minCov']:
#             table[col].append(sum(x * y for x, y in zip(db[col], db['length'])) / sum(db['length']))

#         # Special weighted average
#         db['considered_length'] = [x*y for x,y in zip(db['breadth_minCov'], db['length'])]
#         considered_leng = db['considered_length'].sum()

#         if considered_leng != 0:
#             table['ANI'].append((considered_leng - db['SNPs'].sum()) / considered_leng)
#         else:
#             table['ANI'].append(0)

#         # Special
# #         table['max_cov'].append(db['max_cov'].max())
# #         table['min_cov'].append(db['min_cov'].min())

#     return pd.DataFrame(table)

def _genome_wide_readComparer(gdb, s2b, b2l, **kwargs):
    '''
    Now this does work on the genome level
    '''
    mm_level = kwargs.get('mm_level', False)
    stb = s2b

    if not mm_level:
        gdb = gdb.sort_values('mm').drop_duplicates(
                subset=['scaffold', 'name1', 'name2'], keep='last')\
                .sort_values('scaffold')
        gdb['mm'] = 0
    table = defaultdict(list)

    # Backwards compatibility
    #gdb = gdb.rename(columns={'ANI':'popANI'})

    for mm in sorted(list(gdb['mm'].unique())):
        Odb = gdb[gdb['mm'] <= mm].sort_values('mm').drop_duplicates(subset=['scaffold', 'name1', 'name2'], keep='last')
        for things, db in Odb.groupby(['genome', 'name1', 'name2']):
#         for genome, df in Odb.groupby('genome'):

#     for things, db in gdb.groupby(['genome', 'name1', 'name2', 'mm']):
            genome, name1, name2 = things

            table['genome'].append(genome)
            table['name1'].append(name1)
            table['name2'].append(name2)
            table['mm'].append(mm)

            # Weighted average columns
            tcb = db['compared_bases_count'].sum()
            for col in ['coverage_overlap']:
                if tcb == 0:
                    table[col].append(np.nan)
                else:
                    table[col].append(sum([x * y for x, y in zip(db[col], db['compared_bases_count'])]) / tcb)

            # Sum columns
            for col in ['compared_bases_count', 'consensus_SNPs', 'population_SNPs']:
                if col in list(db.columns):
                    table[col].append(db[col].sum())

            # Special ANI column
            for col in ['ANI', 'popANI', 'conANI']:
                if col in list(db.columns):
                    if tcb == 0:
                        table[col].append(np.nan)
                    else:
                        table[col].append(sum(a * c  if a == a else 0 for a, c in
                                            zip(db[col], db['compared_bases_count'])) / tcb)

            # Percent compared
            if b2l != None:
                table['percent_compared'].append(tcb / b2l[genome])

    db = pd.DataFrame(table)

    if not mm_level:
        del db['mm']

    return db

# def _backfill_blanks(db, s2l):
#     scaffs = list(set(s2l.keys()) - set(db['scaffold'].unique()))
#     bdb = pd.DataFrame({'scaffold':scaffs, 'length':[s2l[s] for s in scaffs]})
#
#     # make some adjustments
#     bdb['bases_w_0_coverage'] = bdb['length']
#
#     # append
#     db = db.append(bdb, sort=True)
#
#     # fill 0
#     return db.fillna(0)

# def interate_sdb_mm(sdb, on='genome', s2l=None, stb=None):
#     '''
#     For the dataframe, iterate through each mm and a dataframe with ALL scaffolds at that mm level
#     (including blanks)
#     '''
#     for g, db in sdb.groupby(on):
#         if s2l == None:
#             gs2l = db.set_index('scaffold')['length'].to_dict()
#         else:
#             gs2l = {s:s2l[s] for s in [x for x, b in stb.items() if b == g]}
#         mms = sorted(db['mm'].unique())
#         for mm in mms:
#             # get all the ones that you can
#             dd = db[db['mm'] <= mm].sort_values('mm').drop_duplicates(subset='scaffold', keep='last')
#             #print("mm={0}; len={1}; tl={2}".format(mm, len(dd), len(s2l.keys())))
#
#             # backfill with blanks
#             dd = _backfill_blanks(dd, gs2l)
#             #print("mm={0}; len={1}; tl={2}".format(mm, len(dd), len(s2l.keys())))
#
#             yield g, mm, dd

def estimate_breadth(coverage):
    '''
    Estimate breadth based on coverage

    Based on the function breadth = -1.000 * e^(0.883 * coverage) + 1.000
    '''
    import numpy as np
    return (-1) * np.exp(-1 * ((0.883) * coverage)) + 1



def _genome():
    return 'genome'

def load_scaff2bin(input_stb, IS=None):
    '''
    From the input, load an .stb. The input can be a lot of things, though
    '''

    # Check if there is nothing there
    if IS is not None:
        if (input_stb == []) | (input_stb is None):
            s2l = IS.get('scaffold2length')
            stb = {}
            for scaffold in list(s2l.keys()):
                stb[scaffold] = 'all_scaffolds'
            logging.info('Scaffold to bin will consider all scaffolds the same genome')
            return stb

    # Check if this is a .fasta file
    try:
        stb = gen_stb(input_stb)
        logging.info('Scaffold to bin was made using .fasta files')
        return stb
    except:
        pass

    # Check if this is a regular stb file
    if (len(input_stb) == 1):
        try:
            stb = parse_stb(input_stb[0])
            logging.info('Scaffold to bin was made using .stb file')
            return stb
        except:
            pass

    # Check if you didn't have an input
    if (input_stb == []):
        return {}

    # Fail
    logging.error('Could not load the scaffold to bin file!')
    assert False

def parse_stb(stb_loc):
    stb = {}
    with open(stb_loc, "r") as ins:
        for line in ins:
            linewords = line.strip().split('\t')
            scaffold,b = linewords[:2]
            scaffold = scaffold.strip()
            b = b.strip()
            if scaffold not in stb:
                stb[scaffold] = []
            stb[scaffold] = b
    return stb

def gen_stb(fastas):
    stb = {}
    for fasta in fastas:
        bin = os.path.basename(fasta)
        for seq_record in SeqIO.parse(fasta, "fasta"):
            id = str(seq_record.id).strip()
            stb[id] = bin
    if len(stb.keys()) == 0:
        raise Exception
    return stb

def calc_bin2scaffols(stb):
    b2s = {}
    for scaffold, bin in stb.items():
        if bin not in b2s:
            b2s[bin] = set()
        b2s[bin].add(scaffold)
    return b2s

def calc_mms(covT):
    '''
    Return a list of mms from scaffold to mm to cov
    '''
    mms = set()
    for scaff, covt in covT.items():
        mms = mms.union(covt.keys())
    return sorted(list(mms))

def generate_genome_coverage_array(covT, s2l, order=None, maxMM=100, mask_edges=0):
    '''
    Geneate a pandas series of coverage values indexed by genome position.

    Arguments:
        covT = The covT object (IS.get('covT')). In this case should have scaffold as well. (scaffold -> mm -> coverage array)
        order = Order of scaffolds to use. Otherwise will just use the sorted order of s2l keys
        s2l = Scaffold 2 length
        maxMM = The maximum number of mms to make it into the final coverage array
        mask_edges = Remove this much from the edge of each scaffold

    Returns:
        cov = Genome-wide array of coverage values
        scaff2index_addition = Dictionary with the property: scaff2index_addition[scaffold] = scalar to sum to scaffold index to get genome_index
    '''
    arrs = []
    tally = 0
    scaff2index_addition = {}

    if order is None:
        order = sorted(list(s2l.keys()), key=s2l.get, reverse=True)

    for scaff in order:
        slen = s2l[scaff]

        if scaff in covT:
            cov = inStrain.profile.profile_utilities.mm_counts_to_counts_shrunk(covT[scaff], maxMM=maxMM, fill_zeros=slen)
        else:
            cov = pd.Series(index=np.arange(slen))


        if mask_edges:
            if len(cov) >= (mask_edges * 2):
                cov = cov[mask_edges:len(cov)-mask_edges]
                slen = slen - (mask_edges * 2)
            else: # really short scaffold
                cov = pd.Series()
                slen = 0

        arrs.append(cov)

        # Set the index addition
        scaff2index_addition[scaff] = tally

        # Update the tally
        tally += slen

    cov = pd.concat(arrs, verify_integrity=False, ignore_index=True).reset_index(drop=True).fillna(0)

    return cov, scaff2index_addition



# Import packages
import os
import gc
import sys
import time
import pysam
import scipy
import psutil
import random
import logging
import resource
import traceback
import numpy as np
import pandas as pd
from tqdm import tqdm
import multiprocessing
import concurrent.futures
from subprocess import call
from concurrent import futures
from collections import defaultdict
from numba import jit

import inStrain.profile.linkage
import inStrain.SNVprofile
import inStrain.readComparer
import inStrain.logUtils
import inStrain.GeneProfile
import inStrain.controller

from inStrain import __version__

'''
The following are globals used during processing
'''
P2C = {'A':0, 'C':1, 'T':2, 'G':3} # base -> position
C2P = {0:'A', 1:'C', 2:'T', 3:'G'} # position -> base

def split_profile_worker(split_cmd_queue, Sprofile_dict, log_list,
                        null_model, bam, single_thread=False):
    '''
    A worker to profile splits that can be multi-processed

    Kind of a wrapper for the function "split_profile_wrapper_groups"

    Args:
        split_cmd_queue: A queue of split commands to grab
        Sprofile_dict: A dictionary to store processed splits
        log_list: A queue to put logs

        null_model: Used for SNP profiling (needs to be pre-generated)
        bam: Location of .bam file

    Keyword Args:
        singe_thread: Run in single_thread mode
    '''
    # Initilize the .bam file
    bam_init = samfile = pysam.AlignmentFile(bam)

    # Apply patch
    inStrain.controller.patch_mp_connection_bpo_17560()

    j = 0
    pid = os.getpid()
    while True:
        j += 1
        group_name = "{0}_{1}".format(pid, j)
        group_log = inStrain.logUtils.get_group_log('SplitProfile', group_name, 'start')

        # Get command
        if not single_thread:
            cmds = split_cmd_queue.get(True)
        else:
            try:
                cmds = split_cmd_queue.get(timeout=5)
            except:
                return

        # Process split
        Splits = split_profile_wrapper_groups(cmds, null_model, bam_init)
        LOG = ''
        for Split in Splits:
            # This is an error
            if type(Split) == type('string'):
                LOG += Split
                continue
            Sprofile_dict[Split.scaffold + '.' + str(Split.split_number)] = Split
            LOG += Split.log + '\n'

        group_log += inStrain.logUtils.get_group_log('SplitProfile', group_name, 'end')
        LOG += group_log + '\n'
        log_list.put(LOG)

def split_profile_wrapper_groups(cmds, null_model, bam_init):
    '''
    A wrapper to profile a list of splits and handle exceptions

    This is really based around the method "profile_split"
    '''
    results = []
    for cmd in cmds:
        try:
            results.append(profile_split(bam_init, cmd.scaffold, cmd.start,
                            cmd.end, cmd.split_number, cmd.sequence, cmd.R2M,
                            null_model, bam_name=cmd.samfile, **cmd.arguments))
        except Exception as e:
            print(e)
            traceback.print_exc()

            t = time.strftime('%m-%d %H:%M')
            log_message = "\n{1} DEBUG FAILURE SplitException {0} {2}\n"\
                            .format(cmd.scaffold, t, cmd.split_number)
            results.append(log_message)
    return results


def profile_split(samfile, scaffold, start, end, split_number,
                    seq, R2M, null_model, **kwargs):
    '''
    The heart of inStrain split profiling

    Args:
        samfile: A pre-initialized pysam.AlignmentFile
        scaffold: The name of the scaffold to profile
        start: Where to start profiling the scaffold (inclusive 0-based)
        end: Where to end profiling the scaffold (inclusive 0-based)
        split_number: Int representation of the split number
        seq: The sequence of the scaffold
        R2M: A dictionary of read -> number of mismatches
        null_model: An initialized null_model

    Returns:
        Sprofile: A SplitObject()
    '''
    # Log
    unit = "{0}.{1}".format(scaffold, split_number)
    log_message = inStrain.logUtils.get_worker_log('SplitProfile', unit, 'start')

    # For testing purposes
    if ((scaffold == 'FailureScaffoldHeaderTesting') & (split_number == 1) & (kwargs.get('debug', False))):
        assert False

    # Get kwargs
    min_cov = int(kwargs.get('min_cov', 5))
    min_covR = int(kwargs.get('rarefied_coverage', 5))
    min_freq = float(kwargs.get('min_freq', .05))
    min_snp = int(kwargs.get('min_snp', 10))
    store_everything = kwargs.get('store_everything', False)

    # Set up the .bam iterator
    try:
        iter = samfile.pileup(scaffold, truncate=True, max_depth=100000,
                                stepper='nofilter', compute_baq=True,
                                ignore_orphans=True, ignore_overlaps=True,
                                min_base_quality=30, start=start, stop=end+1)
    except ValueError:
        logging.error("scaffold {0} is not in the .bam file {1}!".format(scaffold, samfile))
        return None, log_message

    # Initialize
    mLen = len(seq) # Length of sequence
    covT = {} # Dictionary of mm -> positional coverage along the genome
    clonT = {} # Diciontary of mm -> clonality
    clonTR = {} # Diciontary of mm -> rarefied clonality
    p2c = {} # dictionary of position -> cryptic SNPs
    read_to_snvs = defaultdict(_dlist) # Dictionary of mm -> read variant -> count
    snv2mm2counts = {} # dictionary of position to mm to counts for SNP positions
    Stable = defaultdict(list) # Holds SNP information
    if store_everything:
        pileup_counts = np.zeros(shape=(mLen, 4), dtype=int) # Holds all pileup counts - alexcc 5/9/2019
    else:
        pileup_counts = None

    # Do per-site processing on all these objects
    process_bam_sites(scaffold, seq, iter, covT, clonT, clonTR, p2c,
                      read_to_snvs, snv2mm2counts, Stable, pileup_counts,
                      mLen, null_model, R2M, start=start, **kwargs)

    # Shrink these dictionaries into index pandas Series
    covT = shrink_basewise(covT, 'coverage', start=start, len=mLen)
    clonT = shrink_basewise(clonT, 'clonality', start=start, len=mLen)
    clonTR = shrink_basewise(clonTR, 'clonality', start=start, len=mLen)

    # Make the SNP table
    SNPTable = inStrain.profile.snv_utilities.generate_snp_table(Stable, scaffold, p2c)
    if len(SNPTable) > 0:
        SNPTable['position'] = SNPTable['position'] + start

    # Make linkage network and calculate linkage
    mm_to_position_graph = inStrain.profile.linkage.calc_mm_SNV_linkage_network(read_to_snvs, scaff=scaffold)
    LDdb = inStrain.profile.linkage.calculate_ld(mm_to_position_graph, min_snp, snv2mm2counts=snv2mm2counts, scaffold=scaffold)
    if len(LDdb) > 0:
        for p in ['position_A', 'position_B']:
            LDdb[p] = LDdb[p] + start

    # Make a Scaffold profile to return
    Sprofile = SplitObject()
    Sprofile.scaffold = scaffold
    Sprofile.split_number = split_number
    Sprofile.bam = kwargs.get('bam_name')
    Sprofile.length = mLen
    Sprofile.raw_snp_table = SNPTable
    Sprofile.raw_linkage_table = LDdb
    Sprofile.covT = covT
    Sprofile.clonT = clonT
    Sprofile.clonTR = clonTR
    Sprofile.min_freq = min_freq

    # store extra things if required
    if store_everything:
        for att in ['read_to_snvs', 'mm_to_position_graph', 'pileup_counts']:
            setattr(Sprofile, att, eval(att))

    # Return
    log_message += inStrain.logUtils.get_worker_log('SplitProfile', unit, 'end')
    Sprofile.log = log_message

    return Sprofile

def process_bam_sites(scaffold, seq, iter, covT, clonT, clonTR, p2c,
                        read_to_snvs, snv2mm2counts, Stable, pileup_counts,
                        mLen, null_model, R2M, start=0, **kwargs):
    '''
    Iterate through the iterator and fill in the information

    Args:
        scaffold: Name of the scaffold
        seq: Sequence of the scaffold
        iter: The "samfile.pileup" object
        covT, clonT, clonTR, p2c, read_to_snvs, snv2mm2counts, Stable, pileup_counts: Pre-initialized arrays
        mLen: Lnegth of sequence
        R2M: A dictionary of read -> number of mismatches
        null_model: An initialized null_model
    '''
    # Get kwargs
    min_cov = int(kwargs.get('min_cov', 5))
    min_covR = int(kwargs.get('rarefied_coverage', 5))
    min_freq = float(kwargs.get('min_freq', .05))
    min_snp = int(kwargs.get('min_snp', 10))
    store_everything = kwargs.get('store_everything', False)

    for pileupcolumn in iter:
        # NOTE! If you have paired reads in a pileup column, only one will come back starting in pysam v0.15 (which is usually the intended functionality) - M.O. 3.19.19

        TruePosition =  pileupcolumn.pos
        RelPosition = pileupcolumn.pos - start

        # Get basic base counts
        MMcounts = get_base_counts_mm(pileupcolumn, R2M)
        update_covT(covT, MMcounts, RelPosition, mLen)

        # Call SNPs
        snp, bases, total_counts = inStrain.profile.snv_utilities.update_snp_table(
                                    Stable, clonT, clonTR, MMcounts, p2c,
                                    RelPosition, scaffold, mLen, seq[RelPosition],
                                    null_model, min_cov=min_cov, min_covR=min_covR,
                                    min_freq=min_freq)

        # add the counts for this position to the numpy array alexcc
        if store_everything:
            pileup_counts[RelPosition] = total_counts

        # get linked reads
        if snp:
            inStrain.profile.linkage.update_linked_reads(read_to_snvs,
                                    pileupcolumn, MMcounts, RelPosition,
                                    bases, R2M, scaffold=scaffold)
            snv2mm2counts[RelPosition] = MMcounts

def get_base_counts_mm(pileupcolumn, R2M):
    '''
    From a pileupcolumn object, return a dictionary of readMismatches ->
        list with the counts of [A, C, T, G]
    '''
    table = defaultdict(_4zeros)
    for pileupread in pileupcolumn.pileups:
        if not pileupread.is_del and not pileupread.is_refskip:
            try:
                if type(R2M) == type({}):
                    table[R2M[pileupread.alignment.query_name]]\
                    [P2C[pileupread.alignment.query_sequence[pileupread.query_position]]] += 1
                elif pileupread.alignment.query_name in R2M:
                    table[0]\
                    [P2C[pileupread.alignment.query_sequence[pileupread.query_position]]] += 1
            except KeyError: # This would be like an N or something not A/C/T/G
                pass
    return table

def update_covT(covT, MMcounts, position, mLen):
    '''
    Update covT at this position
    '''
    for mm, count in MMcounts.items():
        if mm not in covT:
            covT[mm] = np.zeros(mLen, dtype=int)
        covT[mm][position] = sum(count)

def mm_counts_to_counts(MMcounts, maxMM=100):
    '''
    Take mm counts and return just counts
    '''
    counts = None
    for mm, count in [(mm, count) for mm, count in MMcounts.items() if mm <= maxMM]:
        if counts is None:
            counts = count
        else:
            counts = np.add(counts, count)

    if counts is None:
        return np.zeros(4, dtype=int)

    else:
        return counts

# def mm_counts_to_counts(MMcounts, maxMM=100):
#     '''
#     Take mm counts and return just counts
#     '''
#     counts = np.zeros(4, dtype=int)
#
#     mms = np.array(list(MMcounts.keys()), dtype='int32')
#     covs = np.array(list(MMcounts.values()))
#
#     return _mm_counts_to_counts_fast(mms, covs, counts, maxMM)
#
# @jit(nopython=True)
# def _mm_counts_to_counts_fast(mms, covs, counts, maxMM):
#     '''
#     Fast implementation
#     '''
#     i = 0
#     for mm in mms:
#         if mm <= maxMM:
#             counts = np.add(counts, covs[i])
#         i += 1
#     return counts

def shrink_basewise(mm2array, name, start=0, len=0):
    NAME2TYPE = {'coverage':'int32', 'clonality':'float32', 'snpCounted':'bool'}

    mm2bases = {}
    for mm in list(mm2array.keys()):
        if start != 0:
            db = pd.Series(mm2array[mm], index=pd.RangeIndex(start=start, stop=start+len), dtype=NAME2TYPE[name]).dropna()
        else:
            db = pd.Series(mm2array[mm], dtype=NAME2TYPE[name]).dropna()
        db = db[db > 0]
        mm2bases[mm] = db
        del mm2array[mm]

    return mm2bases

def merge_profile_worker(sprofile_cmd_queue, Sprofile_dict, Sprofiles,
                        null_model, single_thread=False):
    '''
    Worker to merge_splits
    '''
    # Apply patch
    inStrain.controller.patch_mp_connection_bpo_17560()

    j = 0
    pid = os.getpid()
    while True:
        j += 1
        group_name = "{0}_{1}".format(pid, j)
        group_log = inStrain.logUtils.get_group_log('MergeProfile', group_name, 'start')

        if not single_thread:
            cmds = sprofile_cmd_queue.get(True)
        else:
            try:
                cmds = sprofile_cmd_queue.get(timeout=5)
            except:
                return

        last_c = len(cmds) - 1
        result_list = []
        for c, cmd in enumerate(cmds):
            Sprofile = cmd
            Sprofile.null_model = null_model

            for i in range(Sprofile.number_splits):
                Sprofile = Sprofile.update_splits(i, Sprofile_dict.pop(
                                                Sprofile.scaffold + '.' + str(i)))

            Sprofile = Sprofile.merge()

            if Sprofile is not None:
                if Sprofile.profile_genes:
                    try:
                        Sprofile.run_profile_genes()
                    except:
                        t = time.strftime('%m-%d %H:%M')
                        Sprofile.merge_log += "\n{1} DEBUG FAILURE GeneException {0}".format(str(Sprofile.scaffold), t)

            if c == last_c:
                try:
                    group_log += inStrain.logUtils.get_group_log('MergeProfile', group_name, 'end')
                    Sprofile.merge_log += "\n{0}".format(group_log)
                except:
                    pass
            result_list.append(Sprofile)

        Sprofiles.put(result_list)

def merge_basewise(mm2array_list):
    NAME2TYPE = {'coverage':'int32', 'clonality':'float32', 'snpCounted':'bool'}

    mm2bases = {}

    mms = set()
    for mm2bases in mm2array_list:
        mms = mms.union(set(mm2bases.keys()))

    for mm in mms:
        mm2bases[mm] = pd.concat([mm2array[mm] for mm2array in mm2array_list if mm in mm2array], verify_integrity=True)
    return mm2bases

def merge_special(vals, thing):
    #TODO ['read_to_snvs', 'mm_to_position_graph', 'pileup_counts']
    return True

def make_coverage_table(covT, clonT, clonTR, lengt, scaff, SNPTable, null_model,
                        min_freq=0.05, debug=False):
    '''
    Add information to the table
    Args:
        covT: list of coverage values
        lengt: length of scaffold
        scaff: name of scaffold
        Wdb: windows to process over (NOT YET SUPPORTED)
    '''
    table = defaultdict(list)
    for mm in sorted(list(covT.keys())):

        covs = mm_counts_to_counts_shrunk(covT, mm, fill_zeros=int(lengt))

        if len(covs) == 0:
            covs = pd.Series([0]*lengt)

        nonzeros = np.count_nonzero(covs)

        # Get clonalities
        clons = get_basewise_clons(clonT, mm)
        Rclons = get_basewise_clons(clonTR, mm)

        counted_bases = len(clons)
        rarefied_bases = len(Rclons)
        SNS_count, SNV_count, div_site_count, con_snps, pop_snps = \
                    inStrain.profile.snv_utilities.calc_snps(
                                        SNPTable, mm, null_model,
                                        min_freq=min_freq)

        assert len(covs) == lengt, [covs, lengt, mm]

        # fill in all coverage information
        table['scaffold'].append(scaff)
        table['length'].append(lengt)
        table['breadth'].append(nonzeros/lengt)
        table['coverage'].append(np.mean(covs))
        table['coverage_median'].append(int(np.median(covs)))
        table['coverage_std'].append(np.std(covs))
        table['coverage_SEM'].append(scipy.stats.sem(covs))

        if len(clons) > 0:
            mean_c = np.mean(clons)
            median_c = np.median(clons)
            table['nucl_diversity'].append(1-mean_c)
            table['nucl_diversity_median'].append(1-median_c)
        else:
            table['nucl_diversity'].append(np.nan)
            table['nucl_diversity_median'].append(np.nan)

        if len(Rclons) > 0:
            mean_c = np.mean(Rclons)
            median_c = np.median(Rclons)
            table['nucl_diversity_rarefied'].append(1-mean_c)
            table['nucl_diversity_rarefied_median'].append(1-median_c)
        else:
            table['nucl_diversity_rarefied'].append(np.nan)
            table['nucl_diversity_rarefied_median'].append(np.nan)

        table['breadth_minCov'].append(counted_bases / lengt)
        table['breadth_rarefied'].append(rarefied_bases / lengt)
        table['breadth_expected'].append(estimate_breadth(table['coverage'][-1]))

        table['divergent_site_count'].append(div_site_count) # divergent_sites

        table['SNS_count'].append(SNS_count) # SNS_count
        table['SNV_count'].append(SNV_count)

        table['consensus_divergent_sites'].append(con_snps)
        table['population_divergent_sites'].append(pop_snps)

        if counted_bases == 0:
            table['conANI_reference'].append(0)
            table['popANI_reference'].append(0)
        else:
            table['conANI_reference'].append((counted_bases - con_snps)/ counted_bases)
            table['popANI_reference'].append((counted_bases - pop_snps)/ counted_bases)

        table['mm'].append(mm)

    return pd.DataFrame(table)

def mm_counts_to_counts_shrunk(MMcounts, maxMM=100, fill_zeros=False):
    '''
    Converts shunken mm-level items (e.g. covT) to a 1-d array

    Arguments:
        MMcounts: covT or similar object for a single scaffold
        maxMM: The maximum mm level to be allowed into the final array
        fill_zeros: If an int, fill 0s to make the array this length

    Returns:
        A 1-d array of counts

    '''
    if fill_zeros:
        counts = pd.Series(index=np.arange(fill_zeros))
    else:
        counts = pd.Series()

    for mm, count in [(mm, count) for mm, count in MMcounts.items() if mm <= maxMM]:
        counts = counts.add(count, fill_value=0)

    if fill_zeros:
        counts = counts.fillna(0)

    return counts

def get_basewise_clons(clonT, MM, fill_zeros=False):
    p2c = {}
    mms = sorted([int(mm) for mm in list(clonT.keys()) if int(mm) <= int(MM)])
    for mm in mms:
        p2c.update(clonT[mm].to_dict())

    counts = list(p2c.values())

    if fill_zeros:
        counts = counts.append(pd.Series(np.zeros(fill_zeros - len(counts))))

    return counts

def estimate_breadth(coverage):
    '''
    Estimate breadth based on coverage

    Based on the function breadth = -1.000 * e^(0.883 * coverage) + 1.000
    '''
    import numpy as np
    return (-1) * np.exp(-1 * ((0.883) * coverage)) + 1

def _clonT_to_table(clonT):
    '''
    Based on the new clonT

    You want this to have "clonality, mm, and position"

    Called by SNVProfile
    '''
    dbs = []
    for mm, d in clonT.items():
        db = d.to_frame(name='clonality')#pd.DataFrame(d, columns=['clonality'])
        db['mm'] = mm
        db = db.reset_index()
        db = db.rename(columns={'index':'position'})
        dbs.append(db)

    if len(dbs) > 0:
        return pd.concat(dbs)

    else:
        return pd.DataFrame()

def _make_snp_table(Stable):
    '''
    Called by class scaffold_profile().make_cumulative_tables().
    Run on raw_snp_table.
    Very unclear what it actually does
    '''
    if Stable is not False:
        try:
            Sdb = pd.DataFrame(Stable)
            Sdb['scaffold'] = Sdb['scaffold'].astype('category')
            Sdb['con_base'] = Sdb['con_base'].astype('category')
        except KeyError:
            #logging.info("No SNPs detected!")
            Sdb = pd.DataFrame()
    else:
        Sdb = pd.DataFrame()

    return Sdb

def _parse_Sdb(sdb):
    '''
    Add some information to sdb
    '''
    if len(sdb) == 0:
        return sdb

    sdb['var_freq'] = [[a,c,t,g][['A','C','T','G'].index(v)]/s for a,c,t,g,v,s in zip(\
                        sdb['A'], sdb['C'], sdb['T'], sdb['G'], sdb['var_base'], sdb['position_coverage'])]
    sdb['con_freq'] = [[a,c,t,g][['A','C','T','G'].index(v)]/s for a,c,t,g,v,s in zip(\
                        sdb['A'], sdb['C'], sdb['T'], sdb['G'], sdb['con_base'], sdb['position_coverage'])]
    sdb['ref_freq'] = [[a,c,t,g][['A','C','T','G'].index(v)]/s if v in ['A','C','T','G'] else np.nan for a,c,t,g,v,s in zip(\
                        sdb['A'], sdb['C'], sdb['T'], sdb['G'], sdb['ref_base'], sdb['position_coverage'])]

    return sdb

def gen_snv_profile(Sprofiles, ISP_loc=None, **kwargs):
    '''
    Take a bunch of scaffold profiles and make an SNVprofile
    '''
    if ISP_loc == None:
        location = kwargs.get('output', False)
    else:
        location = ISP_loc
        
    store_everything = kwargs.get('store_everything', False)

    # Merge things
    scaffold_list = []
    bam_list = []

    raw_snp_dbs = []
    raw_link_dbs = []

    cumu_scaff_dbs = []
    cumu_snv_dbs = []

    scaffold_2_mm_2_read_2_snvs = {}

    gene_level_results = defaultdict(list)

    for Sprof in Sprofiles:

        scaffold_list.append(Sprof.scaffold)
        bam_list.append(Sprof.bam)

        raw_snp_dbs.append(Sprof.raw_snp_table)
        raw_link_dbs.append(Sprof.raw_linkage_table)

        cumu_scaff_dbs.append(Sprof.cumulative_scaffold_table)
        cumu_snv_dbs.append(Sprof.cumulative_snv_table)

        if hasattr(Sprof, 'mm_reads_to_snvs'):
            scaffold_2_mm_2_read_2_snvs[Sprof.scaffold] = Sprof.mm_reads_to_snvs

        if hasattr(Sprof, 'gene_results'):
            for i, name in enumerate(['coverage', 'clonality', 'SNP_density', 'SNP_mutation_types']):
                gene_level_results[name].append(Sprof.gene_results[i])

    # Make some dataframes
    raw_snp_table = pd.concat(raw_snp_dbs).reset_index(drop=True)
    if len(raw_snp_table) > 0:
        for col in ['A', 'C', 'G', 'T', 'mm', 'position']:
            raw_snp_table[col] = raw_snp_table[col].astype(int)
            raw_snp_table['scaffold'] = raw_snp_table['scaffold'].astype('category')
            raw_snp_table['con_base'] = raw_snp_table['con_base'].astype('category')


    # Make the profile
    Sprofile = inStrain.SNVprofile.SNVprofile(location)

    # Add some things
    Sprofile.store('object_type', 'profile', 'value', 'Type of SNVprofile (profile or compare)')
    Sprofile.store('bam_loc', bam_list[0], 'value', 'Location of .bam file')
    Sprofile.store('scaffold_list', scaffold_list, 'list', '1d list of scaffolds that were profiled')
    Sprofile.store('raw_linkage_table', pd.concat(raw_link_dbs).reset_index(drop=True),
                    'pandas', 'Raw table of linkage information')
    Sprofile.store('raw_snp_table', raw_snp_table,
                    'pandas', 'Contains raw SNP information on a mm level')
    Sprofile.store('cumulative_scaffold_table', pd.concat(cumu_scaff_dbs).reset_index(drop=True),
                    'pandas', 'Cumulative coverage on mm level. Formerly scaffoldTable.csv')
    Sprofile.store('cumulative_snv_table', pd.concat(cumu_snv_dbs).reset_index(drop=True),
                    'pandas', 'Cumulative SNP on mm level. Formerly snpLocations.pickle')

    if scaffold_2_mm_2_read_2_snvs is not {}:
        Sprofile.store('scaffold_2_mm_2_read_2_snvs', scaffold_2_mm_2_read_2_snvs,
                        'pickle', 'crazy nonsense needed for linkage')

    if len(gene_level_results['coverage']) > 0:
        Sprofile.store('genes_coverage', pd.concat(gene_level_results['coverage']).reset_index(drop=True),
                 'pandas', 'Coverage of individual genes')
        Sprofile.store('genes_clonality', pd.concat(gene_level_results['clonality']).reset_index(drop=True),
                 'pandas', 'Clonality of individual genes')
        Sprofile.store('genes_SNP_count', pd.concat(gene_level_results['SNP_density']).reset_index(drop=True),
                 'pandas', 'SNP density and counts of individual genes')
        Sprofile.store('SNP_mutation_types', pd.concat(gene_level_results['SNP_mutation_types']).reset_index(drop=True),
                 'pandas', 'The mutation types of SNPs')

    # Store extra things
    att2descr = {'covT':'Scaffold -> mm -> position based coverage',
                 #'snpsCounted':"Scaffold -> mm -> position based True/False on if a SNPs is there",
                 'clonT':"Scaffold -> mm -> position based clonality",
                 'clonTR':"Scaffold -> mm -> rarefied position based clonality",
                 'read_to_snvs':'?',
                 'mm_to_position_graph':'?'}
    #for att in ['covT', 'snpsCounted', 'clonT', 'read_to_snvs', 'mm_to_position_graph']:
    for att in ['covT', 'clonT', 'clonTR', 'read_to_snvs', 'mm_to_position_graph']:
        if hasattr(Sprofiles[0], att):
            thing = {S.scaffold:getattr(S, att) for S in Sprofiles}
            Sprofile.store(att, thing, 'special', att2descr[att])

    if store_everything:
        counts_table = list() #alexcc 5/9/2019
        for Sprof in Sprofiles:
            counts_table.append(Sprof.pileup_counts) #add pileup counts to list
        counts_table = np.array(counts_table)
        Sprofile.store('counts_table', counts_table, 'numpy', '1d numpy array of 2D counts tables for each scaffold')

    return Sprofile

class ScaffoldSplitObject():
    '''
    This is a temporary object that holds all of the splits in a scaffold, and can eventually merge them into a scaffold_profile
    '''
    def __init__(self, number_splits):
        '''
        Initialize based on the number of expected splits to calculate
        '''
        self.number_splits = int(number_splits)
        self.split_dict = {i:None for i in range(number_splits)}
        assert self.number_splits == len(self.split_dict.keys())

    def ready(self):
        '''
        Return True if all splits are calculated
        '''
        calced = sum([True if v is not None else False for n, v in self.split_dict.items()])
        return calced == self.number_splits

    def print_status(self):
        '''
        Print a status report
        '''
        calced = sum([True if v is not None else False for n, v in self.split_dict.items()])
        print("{0} splits; {1} done".format(self.number_splits, calced))

    def update_splits(self, number, obj):
        '''
        Return a copy of the object with the split_dict updated
        '''
        self.split_dict[number] = obj
        return self

    def merge(self):
        '''
        Merge the split dict in order to make a scaffold profile
        '''
        log_message = inStrain.logUtils.get_worker_log('MergeProfile', self.scaffold, 'start')

        try:
            if self.number_splits == 1:
                Sprofile = self.split_dict[0].merge_single_profile(self)
                log_message += inStrain.logUtils.get_worker_log('MergeProfile', self.scaffold, 'end')
                Sprofile.merge_log = log_message
                return Sprofile

            else:
                Sprofile = scaffold_profile()
                Sprofile.null_model = self.null_model
                # Handle gene stuff
                for att in ['profile_genes', 'gene_database', 'gene2sequence']:
                    if hasattr(self, att):
                        setattr(Sprofile, att, getattr(self, att))

                # Handle value objects
                for att in ['scaffold', 'bam', 'min_freq']:
                    vals = set([getattr(Split, att) for num, Split in self.split_dict.items()])
                    assert len(vals) == 1, vals
                    setattr(Sprofile, att, list(vals)[0])

                # Handle sum objects
                for att in ['length']:
                    vals = [getattr(Split, att) for num, Split in self.split_dict.items()]
                    setattr(Sprofile, att, sum(vals))

                # Handle dataframes
                for att in ['raw_snp_table', 'raw_linkage_table']:
                    val = pd.concat([getattr(Split, att) for num, Split in self.split_dict.items()]).reset_index(drop=True)
                    setattr(Sprofile, att, val)

                # Handle series
                for att in ['covT', 'clonT', 'clonTR']:
                    vals = [getattr(Split, att) for num, Split in self.split_dict.items()]
                    setattr(Sprofile, att, merge_basewise(vals))

                # Handle extra things
                for att in ['read_to_snvs', 'mm_to_position_graph', 'pileup_counts']:
                    if hasattr(self.split_dict[0], att):
                        vals = [getattr(Split, att) for num, Split in self.split_dict.items()]
                        setattr(Sprofile, att, merge_special(vals, att))

                Sprofile.make_cumulative_tables()

                log_message += inStrain.logUtils.get_worker_log('MergeProfile', self.scaffold, 'end')
                Sprofile.merge_log = log_message

                return Sprofile
        except:
            traceback.print_exc()

            t = time.strftime('%m-%d %H:%M')
            log_message = "\n{1} DEBUG FAILURE MergeError {0}\n"\
                            .format(self.scaffold, t)
            print(log_message)

            return None

    def delete_self(self):
        '''
        Delete everything to free up RAM
        '''
        pass


class SplitObject():
    '''
    Holds the profile of an individual split
    '''
    def __init__(self):
        pass

    def merge_single_profile(self, ScaffoldSplitObject):
        '''
        Convert self into scaffold_profile object
        '''
        Sprofile = scaffold_profile()
        Sprofile.scaffold = self.scaffold
        Sprofile.null_model = ScaffoldSplitObject.null_model
        Sprofile.bam = self.bam
        Sprofile.length = self.length
        Sprofile.raw_snp_table = self.raw_snp_table
        Sprofile.raw_linkage_table = self.raw_linkage_table
        Sprofile.covT = self.covT
        Sprofile.clonT = self.clonT
        Sprofile.clonTR = self.clonTR
        Sprofile.min_freq = self.min_freq

        for att in ['read_to_snvs', 'mm_to_position_graph', 'pileup_counts'] + ['profile_genes', 'gene_database', 'gene2sequence']:
            if hasattr(self, att):
                setattr(Sprofile, att, getattr(self, att))

        for att in ['profile_genes', 'gene_database', 'gene2sequence']:
            if hasattr(ScaffoldSplitObject, att):
                setattr(Sprofile, att, getattr(ScaffoldSplitObject, att))

        # Make cummulative tables
        Sprofile.make_cumulative_tables()

        # Return
        return Sprofile

class scaffold_profile():
    '''
    This is a temporary object that represents the profile of a single scaffold
    '''
    def __init__(self, **kwargs):
        '''
        initialize
        '''
        self.version = __version__

    def make_cumulative_tables(self):
        '''
        Make cumulative tables (still raw-looking)
        This is all on the scaffold level
        '''
        if (self.raw_snp_table is not None):
            self.cumulative_snv_table = _make_snp_table(self.raw_snp_table)
            self.cumulative_snv_table = _parse_Sdb(self.cumulative_snv_table)

        self.cumulative_scaffold_table = make_coverage_table(self.covT, self.clonT,
                            self.clonTR, self.length, self.scaffold, self.raw_snp_table,
                            self.null_model, min_freq=self.min_freq)

    def run_profile_genes(self):
        """
        Run gene-level profile
        """
        if hasattr(self, 'gene_database'):

            scaffold = self.scaffold
            Gdb = self.gene_database
            covT = self.covT
            clonT = self.clonT
            gene2sequence = self.gene2sequence
            cumulative_snv_table = self.cumulative_snv_table

            self.gene_results = inStrain.GeneProfile.profile_genes_from_profile(scaffold, Gdb, covT, clonT, cumulative_snv_table, gene2sequence)

def _dlist():
    return defaultdict(list)

def _4zeros():
    return np.zeros(4, dtype=int)



#!/usr/bin/env python
"""
Run tests
"""

import warnings

import pandas as pd

import tests.test_compare
import tests.test_filter_reads
import tests.test_genes
import tests.test_genome_level
import tests.test_plotting
import tests.test_profile
import tests.test_quick_profile
import tests.test_snv_profile
import tests.test_special

warnings.filterwarnings("ignore")

pd.set_option('display.max_columns', None)
pd.set_option('display.max_colwidth', -1)

# This is from test_compare.py
def UPDATE_COMPARE_TEST_DATA(self):
    """
    Run inStrain on bam1 and bam2, and store the results where IS1 and IS2 are
    """
    # Run the first one
    base1 = self.test_dir + os.path.basename(self.IS1)
    cmd = "inStrain profile {0} {1} -o {2} --skip_plot_generation".format(self.bam1, self.fasta,
                                                                          base1)
    print(cmd)
    code = call(cmd, shell=True)
    assert code == 0, code

    # Copy to new location
    if os.path.isdir(self.IS1):
        shutil.rmtree(self.IS1)
    shutil.copytree(base1, self.IS1)

    # Run the second one
    base2 = self.test_dir + os.path.basename(self.IS2)
    cmd = "inStrain profile {0} {1} -o {2} --skip_plot_generation".format(self.bam2, self.fasta,
                                                                          base2)
    print(cmd)
    code = call(cmd, shell=True)
    assert code == 0, code

    # Copy to new location
    if os.path.isdir(self.IS2):
        shutil.rmtree(self.IS2)
    shutil.copytree(base2, self.IS2)

if __name__ == '__main__':
    tests.test_compare.test_readcomparer().run(tests=['18'], cleanUp=True)

    # tests.test_profile.test_profile().run(tests=[16], cleanUp=False)
    # tests.test_compare.test_readcomparer().run(tests=[13], cleanUp=False)

    # tests.test_profile.test_profile().run()
    # tests.test_filter_reads.TestFilterReads().run()
    # tests.test_snv_profile.test_SNVprofile().run()
    # tests.test_genes.test_gene_statistics().run()
    # tests.test_quick_profile.test_quickProfile().run()
    # tests.test_genome_level.test_genome_wide().run()
    # tests.test_plotting.test_plot().run()
    # tests.test_compare.test_readcomparer().run()
    # tests.tests_special.test_special().run()

    print('everything is working swimmingly!')



import logging
import traceback
import pandas as pd
import seaborn as sns

from collections import defaultdict
import numpy as np

import matplotlib

matplotlib.use('Agg')
matplotlib.rcParams['pdf.fonttype'] = 42

import matplotlib.ticker as ticker
from matplotlib import pyplot as plt
import matplotlib.patches as mpatches
from matplotlib.backends.backend_pdf import PdfPages

import inStrain.plotting.utilities
from inStrain.plotting.utilities import plot_genome
from inStrain.plotting.utilities import estimate_breadth

def allele_freq_plot_from_IS(IS, plot_dir=False, **kwargs):
    # Load the required data
    try:
        if not hasattr(sns, 'histplot') and callable(getattr(sns, 'histplot')):
            raise Exception("Cannot make plot 4 because your seaborn is out of date- need v0.11+")

        db = IS.get('cumulative_snv_table')
        if len(db) == 0:
            return
        db = db.sort_values('mm').drop_duplicates(subset=['scaffold', 'position'], keep='last')\
                    .sort_index().drop(columns=['mm'])
        db = db[(db['cryptic'] == False)]
        if 'allele_count' in db.columns:
            db = db[db['allele_count'] >= 2]
        if 'morphia' in db.columns:
            db = db[db['morphia'] >= 2]

        stb = IS.get('scaffold2bin')
        Mdb = inStrain.genomeUtilities._add_stb(db, stb, verbose=False)
        assert len(Mdb) > 0
    except:
        logging.error("Skipping plot 4 - you don't have all required information. You need to run inStrain genome_wide first")
        traceback.print_exc()
        return

    # Make the plot
    logging.info("Plotting plot 4")
    name = 'MajorAllele_frequency_plot.pdf'
    pp = PdfPages(plot_dir + name)

    for genome, mdb in Mdb.groupby('genome'):
        if not plot_genome(genome, IS, **kwargs):
            continue
        db = major_allele_freq_plot(mdb, title=genome)
        fig = plt.gcf()
        fig.set_size_inches(6, 4)
        fig.tight_layout()
        pp.savefig(fig)#, bbox_inches='tight')
        #plt.show()
        plt.close(fig)

    # Save the figure
    pp.close()
    #plt.show()
    plt.close('all')

def major_allele_freq_plot(db, title=''):
    db.loc[:, 'major_allele_freq'] = [max(x, y) for x, y in zip(db['var_freq'], db['ref_freq'])]
    sns.histplot(db['major_allele_freq'], bins=np.arange(0.5, 1, 0.01), kde=False, binwidth=0.005)

    plt.xlim(0.5, 1)
    plt.title(title)
    plt.xlabel('Major allele frequency')
    plt.ylabel('Number of SNPs')


# Get the version
from ._version import __version__

# Import packages
import gc
import os
import sys
import h5py
import copy
import pysam
import logging
import argparse
import pandas as pd
pd.set_option('mode.chained_assignment',None)
import functools
import logging
import struct
import sys

from subprocess import call
from datetime import datetime
from collections import defaultdict

# Import inStrain stuff
import inStrain.profile
import inStrain.profile.samtools_ops

import inStrain.plotting
import inStrain.plotting.plotting_controller

import inStrain.filter_reads
import inStrain.utils
import inStrain.readComparer
import inStrain.GeneProfile
import inStrain.genomeUtilities
import inStrain.quickProfile
import inStrain.SNVprofile
import inStrain.logUtils
import inStrain.compare_controller

class Controller():
    '''
    Controller of the whole shebang
    '''

    def main(self, args):
        ''' Parse user options and call the correct pipeline'''
        # Patch multiprocessing
        patch_mp_connection_bpo_17560()

        # Call the appropriate workflow
        if args.operation == "profile":
            self.profile_operation(args)

        if args.operation == "compare":
            self.compare_operation(args)

        if args.operation == "filter_reads":
            self.filter_reads_operation(args)

        if args.operation == "profile_genes":
            self.profile_genes_operation(args)

        if args.operation == "genome_wide":
            self.genome_wide_operation(args)

        if args.operation == "quick_profile":
            self.quick_profile_operation(args)

        if args.operation == "plot":
            self.plot_operation(args)

        if args.operation == "other":
            self.other_operation(args)

        if args.operation == "check_deps":
            self.checkdep_operation(args)

        self.shutdown(args)

    def profile_operation(self, args):
        ProfileController(args).main()

    def compare_operation(self, args):
        inStrain.compare_controller.CompareController(args).main()

    def filter_reads_operation(self, args):
        inStrain.filter_reads.Controller().main(args)

    def profile_genes_operation(self, args):
        inStrain.GeneProfile.Controller().main(args)

    def genome_wide_operation(self, args):
        inStrain.genomeUtilities.Controller().main(args)

    def quick_profile_operation(self, args):
        inStrain.quickProfile.main(args)

    def plot_operation(self, args):
        inStrain.plotting.plotting_controller.PlottingController(args).main()
        #inStrain.plottingUtilities.main(args)

    def other_operation(self, args):
        # Check if you should convert IS profile
        if args.old_IS != None:
            inStrain.SNVprofile.convert_SNVprofile(args.old_IS)
        if args.run_statistics != None:
            inStrain.logUtils.process_logs(args.run_statistics)

    def checkdep_operation(self, args):
        message = inStrain.utils.gen_dependency_report()
        print(message)

    def shutdown(self, args):
        try:
            logloc = logging.getLoggerClass().root.handlers[0].baseFilename
        except:
            return
        logging.debug("inStrain complete; shutting down logger and printing run stats (log location = {0})".format(logloc))
        logging.shutdown()
        inStrain.logUtils.report_run_stats(logloc, most_recent=True, printToo=args.debug, debug=args.debug)

class ProfileController(object):
    '''
    Main controller of the profile command
    '''
    def __init__(self, args):
        '''
        Set all of the command line arguments in the "args" attribute

        Doing it this way lets your pass the arguments to other controllers
        '''
        self.args = args
        self.ori_args = copy.deepcopy(args)

    def main(self):
        '''
        The main method when run on the command line
        '''
        # Parse arguments
        self.validate_arguments()

        # Filter reads
        self.profile_filter_reads()

        # Profile
        self.run_profile()

        # Profile genes
        # self.profile_profile_genes()

        # Make things genome_wide
        self.profile_genome_wide()

        # Make plots
        self.profile_plots()

        # Final message
        self.write_final_message()

        return self.ISP

    def validate_arguments(self):
        '''
        Do some parsing, start up a logger
        '''
        # Get out the "args" to manipulate it
        args = self.args

        # default prefix is now fasta prefix -alexcc 5/8/2019
        if args.output == 'inStrain':
            args.output = args.fasta.split(".")[0].split("/")[-1]

        # Set up "base"
        out_base = args.output

        # Set up Logger
        outbase = out_base
        ISP = inStrain.SNVprofile.SNVprofile(outbase)
        log_loc = ISP.get_location('log') + 'log.log'
        setup_logger(log_loc)

        # Make the bam file if you need to; remove it from args
        self.bam = inStrain.profile.samtools_ops.prepare_bam_fie(args.bam, args.processes)
        del self.args.bam

        # Load the list of scaffolds
        args.scaffolds_to_profile = inStrain.profile.fasta.load_scaff_list(
                                                    args.scaffolds_to_profile)

        # Fix the fdr
        if args.fdr == 0:
            args.fdr = 1e-6

        # Handle database mode
        if args.database_mode:
            args.min_read_ani = 0.92
            args.skip_mm_profiling = True
            args.min_genome_coverage = 1

        # Make sure you have a .stb if needed
        if args.min_genome_coverage != 0:
            assert args.stb != [], 'If you adjust the minimum genome coverage, you need to provide an .stb!'

        # See if you have genes
        if args.gene_file != None:
            self.genes_file_present = True

        self.args = args
        self.args.IS = ISP.get('location')
        self.ISP = ISP
        self.kwargs = vars(self.args)

    def profile_filter_reads(self):
        '''
        Call the filter reads module as run with "profile"
        '''
        message = """\
***************************************************
    ..:: inStrain profile Step 1. Filter reads ::..
***************************************************
        """
        logging.info(message)
        inStrain.logUtils.log_checkpoint("main_profile", "filter_reads", "start")

        # Profile reads
        Rdic, RR, fasta_db, scaff2sequence, s2l = \
                inStrain.filter_reads.Controller().main_from_profile(self.ISP,
                                                    self.bam, **self.kwargs)

        # Store results
        self.RR = RR
        self.Rdic = Rdic
        self.fasta_db = fasta_db
        self.scaff2sequence = scaff2sequence
        self.scaffold2length = s2l

        # Parse results
        inStrain.logUtils.log_checkpoint("FilterReads", "parse_results", "start")
        self.parse_filter_reads()
        inStrain.logUtils.log_checkpoint("FilterReads", "parse_results", "end")

        inStrain.logUtils.log_checkpoint("main_profile", "filter_reads", "end")

    def parse_filter_reads(self):
        '''
        Parse filter read results in the context of profile
        '''
        # Parse results
        self.scaffold2pairs = self.RR.set_index('scaffold')['filtered_pairs'].to_dict()
        self.readlength = float(self.RR.loc[0, 'mean_pair_length'])

        # Filter the .fasta file with these results
        self.fasta_db = inStrain.profile.fasta.filter_fasta(self.fasta_db,
                        self.scaffold2pairs, self.scaffold2length,
                        self.readlength, **self.kwargs)

        # Change up Rdic if needed
        if self.args.skip_mm_profiling:
            newRdic = {}
            for s, p2i in self.Rdic.items():
                newRdic[s] = set(p2i.keys())
            self.Rdic = newRdic
            self.ISP.store('Rdic', self.Rdic, 'pickle', 'list of filtered read pairs')
        else:
            self.ISP.store('Rdic', self.Rdic, 'dictionary', 'Read pair -> mismatches')


        # Store results
        self.ISP.store('mapping_info', self.RR, 'pandas', "Report on reads")
        self.ISP.store('fasta_loc', os.path.abspath(self.args.fasta), 'value', 'Location of .fasta file used during profile')
        self.ISP.store('scaffold2length', self.scaffold2length, 'dictionary', 'Dictionary of scaffold 2 length')

        # Print status report
        unfiltered_pairs = self.RR['unfiltered_pairs'].iloc[0]
        filterd_pairs = self.RR['filtered_pairs'].iloc[0]
        pair_length = self.readlength

        status = ''
        status += "{0:.1f}% of reads were removed during filtering\n".format(
                    ((unfiltered_pairs - filterd_pairs) / unfiltered_pairs)*100)
        status += "{0:,} read pairs remain ({1:#.4g} Gbp)".format(
                    filterd_pairs, (filterd_pairs * pair_length)/1e9)
        logging.info(status)

        # Handle exceptions
        if self.RR['filtered_pairs'].tolist()[0] == 0:
            logging.error(
                "Because no read pairs remain I'm going to crash now. Maybe this is failing because you dont have paired reads (in which case you should adjust --pairing_filter option), or maybe its failing because the mapper you used uses full fasta headers (in which case you should use the flag --use_full_fasta_header)")
            raise Exception('No paired reads detected; see above message and log')

        if len(self.fasta_db) <= 0:
            logging.error("No scaffolds passed initial filtering based on numbers of mapped reads")
            raise Exception('No scaffolds detected; see above message and log')

    def run_profile(self):
        '''
        Call the actual profile module
        '''
        message = """\
***************************************************
.:: inStrain profile Step 2. Profile scaffolds ::..
***************************************************
        """
        logging.info(message)
        inStrain.logUtils.log_checkpoint("main_profile", "profile_scaffolds", "start")

        # Do some argument handling
        self.kwargs['s2s'] = self.scaff2sequence
        self.kwargs['s2p'] = self.scaffold2pairs

        # Call the module
        self.ISP = inStrain.profile.profile_bam(self.bam, self.fasta_db,
                                                self.Rdic, self.ISP.get('location'),
                                                **self.kwargs)

        # Write output
        inStrain.logUtils.log_checkpoint("Profile", "store_output", "start")
        self.write_output()
        inStrain.logUtils.log_checkpoint("Profile", "store_output", "end")

        inStrain.logUtils.log_checkpoint("main_profile", "profile_scaffolds", "end")

    def write_output(self):
        '''
        Write output files
        '''
        logging.debug("Writing output files now")

        for t in ['SNVs', 'scaffold_info', 'SNVs', 'linkage', 'gene_info']:
            self.ISP.generate(t)
        self.ISP.generate('mapping_info', **self.kwargs)

    def profile_profile_genes(self):
        '''
        Call profile genes from the "profile" module
        '''
        message = """\
***************************************************
  .:: inStrain profile Step 3. Profile genes ::..
***************************************************
        """
        logging.info(message)
        args = self.args

        if args.gene_file != None:
            inStrain.logUtils.log_checkpoint("main_profile", "profile_genes", "start")
            Controller().profile_genes_operation(copy.deepcopy(args))
            inStrain.logUtils.log_checkpoint("main_profile", "profile_genes", "end")
        else:
            logging.info('Nevermind! You didnt include a genes file')

    def profile_genome_wide(self):
        '''
        Call genome_wide from "profile" module
        '''
        message = """\
***************************************************
.:: inStrain profile Step 4. Make genome-wide ::..
***************************************************
        """
        logging.info(message)
        args = self.args

        if not args.skip_genome_wide:
            inStrain.logUtils.log_checkpoint("main_profile", "genome_wide", "start")
            Controller().genome_wide_operation(copy.deepcopy(args))
            inStrain.logUtils.log_checkpoint("main_profile", "genome_wide", "end")
        else:
            logging.info('Nevermind! You chose to skip genome_wide')

    def profile_plots(self):
        '''
        Call plotting function from "profile" module
        '''
        # Generate plots
        message = """\
***************************************************
 .:: inStrain profile Step 5. Generate plots ::..
***************************************************
        """
        logging.info(message)
        args = self.args

        if not args.skip_plot_generation:
            inStrain.logUtils.log_checkpoint("main_profile", "making_plots", "start")
            args.plots = 'a'
            Controller().plot_operation(args)
            inStrain.logUtils.log_checkpoint("main_profile", "making_plots", "end")
        else:
            logging.info('Nevermind! You chose to skip making plots')

    def write_final_message(self):
        Sprofile = self.ISP
        message = """\
$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

..:: inStrain profile finished ::..

Output tables........ {0}
Figures.............. {1}
Logging.............. {2}

See documentation for output descriptions - https://instrain.readthedocs.io/en/latest/

$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
        """.format(Sprofile.get_location('output'), \
            Sprofile.get_location('figures'),
            Sprofile.get_location('log'))
        logging.info(message)

def setup_logger(loc):
    ''' set up logger such that DEBUG goes only to file, rest go to file and console '''

    # Cancel if a logger already exists:
    if logging.getLogger('').handlers:
        return

    # set up logging everything to file
    logging.basicConfig(level=logging.DEBUG,
                       format='%(asctime)s %(levelname)-8s %(message)s',
                       datefmt='%y-%m-%d %H:%M:%S',
                       filename=loc)

    # set up logging of INFO or higher to sys.stderr
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    formatter = logging.Formatter('%(message)s')
    console.setFormatter(formatter)

    logging.getLogger('').addHandler(console)

    logging.debug("!"*80)
    logging.debug("***Logger started up at {0}***".format(loc))
    logging.debug("Command to run inStrain was: {0}\n".format(' '.join(sys.argv)))
    logging.debug("inStrain version {0} was run \n".format(__version__))
    logging.debug("!"*80 + '\n')

    try:
        logging.debug(inStrain.utils.gen_dependency_report())
    except:
        logging.warning('Failed to log dependencies')


def patch_mp_connection_bpo_17560():
    """
    From https://stackoverflow.com/questions/47776486/python-struct-error-i-format-requires-2147483648-number-2147483647

    Apply PR-10305 / bpo-17560 connection send/receive max size update

    See the original issue at https://bugs.python.org/issue17560 and
    https://github.com/python/cpython/pull/10305 for the pull request.

    This only supports Python versions 3.3 - 3.7, this function
    does nothing for Python versions outside of that range.

    """
    patchname = "Multiprocessing connection patch for bpo-17560"
    if not (3, 3) < sys.version_info < (3, 8):
        # print(
        #     patchname + " not applied, not an applicable Python version: %s",
        #     sys.version
        # )
        return

    from multiprocessing.connection import Connection

    orig_send_bytes = Connection._send_bytes
    orig_recv_bytes = Connection._recv_bytes
    if (
        orig_send_bytes.__code__.co_filename == __file__
        and orig_recv_bytes.__code__.co_filename == __file__
    ):
        #print(patchname + " already applied, skipping")
        return

    @functools.wraps(orig_send_bytes)
    def send_bytes(self, buf):
        n = len(buf)
        if n > 0x7fffffff:
            pre_header = struct.pack("!i", -1)
            header = struct.pack("!Q", n)
            self._send(pre_header)
            self._send(header)
            self._send(buf)
        else:
            orig_send_bytes(self, buf)

    @functools.wraps(orig_recv_bytes)
    def recv_bytes(self, maxsize=None):
        buf = self._recv(4)
        size, = struct.unpack("!i", buf.getvalue())
        if size == -1:
            buf = self._recv(8)
            size, = struct.unpack("!Q", buf.getvalue())
        if maxsize is not None and size > maxsize:
            return None
        return self._recv(size)

    Connection._send_bytes = send_bytes
    Connection._recv_bytes = recv_bytes

   # print(patchname + " applied")



import logging
import traceback
import pandas as pd
import seaborn as sns

from collections import defaultdict
import numpy as np

import matplotlib
import warnings
import scipy
import scipy.spatial.distance as ssd

matplotlib.use('Agg')
matplotlib.rcParams['pdf.fonttype'] = 42

import matplotlib.ticker as ticker
from matplotlib import pyplot as plt
import matplotlib.patches as mpatches
from matplotlib.backends.backend_pdf import PdfPages

import inStrain.compare_utils
import inStrain.plotting.utilities
from inStrain.plotting.utilities import plot_genome
from inStrain.plotting.utilities import estimate_breadth

def dendrograms_from_RC(IS, plot_dir=False, **kwargs):
    # Load the required data
    try:
        db = IS.get('comparisonsTable')
        stb = IS.get('scaffold2bin')
        b2l = IS.get('bin2length')
        gdb = inStrain.genomeUtilities._add_stb(db, stb)
        Mdb = inStrain.genomeUtilities._genome_wide_readComparer(gdb, stb, b2l, mm_level=False)

        Mdb['name1'] = [inStrain.plotting.utilities._shorten_name(x) for x in Mdb['name1']]
        Mdb['name2'] = [inStrain.plotting.utilities._shorten_name(x) for x in Mdb['name2']]

        Mdb = Mdb.sort_values(['genome', 'name1', 'name2'])
        assert len(Mdb) > 0
    except:
        logging.error("Skipping plot 10 - you don't have all required information. You need to run inStrain genome_wide first")
        traceback.print_exc()
        return

    # Plot
    logging.info("Plotting plot 10")
    name = 'inStrainCompare_dendrograms.pdf'
    pp = PdfPages(plot_dir + name)

    for genome, mdb in Mdb.groupby('genome'):
        # if not plot_genome(genome, IS, **kwargs):
        #     continue
        plot_readComparerer_dendrograms(mdb, genome, cluster_method='average')
        fig = plt.gcf()
        #fig.tight_layout()
        pp.savefig(fig, bbox_inches='tight')
        plt.close(fig)

    # Save the figure
    pp.close()
    #plt.show()
    plt.close('all')

    print('Done!')

def plot_readComparerer_dendrograms(gdb, title, cluster_method='single', thresh=0.001, gthresh = 0.01):
    # Set up the dataframe
    gdb = add_av_RC(gdb)

    # Set up the figure
    fig, (ax1, ax2) = plt.subplots(1, 2, gridspec_kw={'wspace': 0.5})

    #for thing in ['av_cov', 'av_ani']:
    for thing in ['av_ani', 'av_cov']:
        t2a = {'av_ani':ax1, 'av_cov':ax2}
        t2c = {'av_ani':'Average Nucleotide Identity (ANI)', 'av_cov':'Shared Genome Coverage (%)'}
        t2t = {'av_ani':thresh, 'av_cov':gthresh}

        gdb['dist'] = 1 - gdb[thing]
        db = gdb.pivot("name1", "name2", 'dist')

        names = db.columns
        arr =  np.asarray(db)
        arr = ssd.squareform(arr, checks=True)

        # Cluster
        linkage = scipy.cluster.hierarchy.linkage(arr, method=cluster_method)
        fclust = scipy.cluster.hierarchy.fcluster(linkage, thresh,
                            criterion='distance')
        Cdb = inStrain.compare_utils._gen_cdb_from_fclust(fclust, names)

        if thing == 'av_ani':
            name2cluster = Cdb.set_index('genome')['cluster'].to_dict()
            name2color = inStrain.plotting.utilities.gen_color_dictionary(names, name2cluster)

        # Plot
        plt.sca(t2a[thing])
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            _make_RC_dendrogram(linkage, names, xlabel=t2c[thing], subtitle_string=title, threshold=t2t[thing],
                               name2color=name2color)

    # Fix axis labels 1
    for thing in ['av_cov', 'av_ani']:
        axes = t2a[thing]
        labels = axes.xaxis.get_majorticklocs()
        for i, label in enumerate(labels):
            labels[i] = round((1 - float(label)) * 100, 3)
        axes.set_xticklabels(labels)

def add_av_RC(db):
    '''
    add a column titled 'av_ani' to the passed in dataframe
    dataframe must have rows reference, querey, and ani
    Args:
        db: dataframe
    '''

    combo2value = defaultdict(lambda: np.nan)
    combo2value2 = defaultdict(lambda: np.nan)
    for i, row in db.iterrows():
        combo2value["{0}-vs-{1}".format(row['name1'], row['name2'])] \
            = row['popANI']
        combo2value2["{0}-vs-{1}".format(row['name1'], row['name2'])] \
            = row['coverage_overlap']

    table = defaultdict(list)
    samples = set(db['name1'].tolist()).union(set(db['name2'].tolist()))
    for samp1 in samples:
        for samp2 in samples:
            if samp1 == samp2:
                table['name1'].append(samp1)
                table['name2'].append(samp2)
                table['av_ani'].append(1)
                table['av_cov'].append(1)

            else:
                table['name1'].append(samp1)
                table['name2'].append(samp2)
                table['av_ani'].append(np.nanmean([combo2value["{0}-vs-{1}".format(samp1,samp2)], \
                            combo2value["{0}-vs-{1}".format(samp2,samp1)]]))
                table['av_cov'].append(np.nanmean([combo2value2["{0}-vs-{1}".format(samp1,samp2)], \
                            combo2value2["{0}-vs-{1}".format(samp2,samp1)]]))

    return pd.DataFrame(table)

def _make_RC_dendrogram(linkage, names, **kwargs):
    '''
    Make the dendrogram used in plot 2
    names can be gotten like:
        db = db.pivot("reference","querry","ani")
        names = list(db.columns)
    Args:
        linkage: result of scipy.cluster.hierarchy.linkage
        names: names of the linkage
    Kwargs:
        name2cluster: dict
        self_thresh: x-axis for soft line
        threshold: x-axis for hard line
        title_sting: title of the plot
        subtitle_string: subtitle of the plot
        winners: list of "winning" genomes (to be marked with star)
        genome2taxonomy: dictionary to add taxonomy information
    Returns:
        Matplotlib primed with a figure
    '''
    # Load possible kwargs
    #name2cluster = kwargs.get('name2cluster',False)
    name2color = kwargs.get('name2color',False)
    self_thresh = kwargs.get('self_thresh',False)
    threshold = kwargs.get('threshold',False)
    threshold = False
    title_string = kwargs.get('title_string','')
    subtitle_string = kwargs.get('subtitle_string','')
    winners = kwargs.get('winners',False)
    genome2taxonomy = kwargs.get('genome2taxonomy',False)
    xlabel = kwargs.get('xlabel','Average Nucleotide Identity (ANI)')

#     # Make special things
#     if name2cluster != False:
#         name2color = drep.d_analyze.gen_color_dictionary(names, name2cluster)
#     else:
#         name2color = False

    # Make the dendrogram
    sns.set_style('whitegrid')
    g = fancy_dendrogram(linkage,names,name2color,threshold=threshold,self_thresh =\
                        self_thresh)

    # Add the title and subtitle
    plt.suptitle(title_string, y=1, fontsize=18)
    plt.title(subtitle_string, fontsize=10)

#     # Adjust the x-axis
    plt.xlabel(xlabel)
    if threshold != False:
        plt.xlim([0,3*threshold])
    plt.tick_params(axis='x', which='major', labelsize=8)
    plt.tick_params(axis='y', which='major', labelsize=12)
#     axes = plt.gca()
#     labels = axes.xaxis.get_majorticklocs()
#     for i, label in enumerate(labels):
#         labels[i] = round((1 - float(label)) * 100, 2)
#     axes.set_xticklabels(labels)
    plt.gca().yaxis.grid(False)

    # Adjust the figure size
    fig = plt.gcf()
    fig.set_size_inches(20,inStrain.plotting.utilities._x_fig_size(len(names),factor=.5))
    plt.subplots_adjust(left=0.5)

    # Mark winning ones
    if type(winners) is not bool:
        ax = plt.gca()
        labels = [item.get_text() for item in ax.get_yticklabels()]
        for i, label in enumerate(labels):
            if label in winners: labels[i] = label + ' *'
        ax.set_yticklabels(labels)

    # Add taxonomy
    if genome2taxonomy != False:
        g2t = genome2taxonomy
        axes = plt.gca()
        labels = [item.get_text() for item in axes.get_yticklabels()]
        for i, label in enumerate(labels):
            labels[i] = "{0}\n{1}".format(label, g2t[label.replace(' *','')])
        axes.set_yticklabels(labels)

def fancy_dendrogram(linkage,names,name2color=False,threshold=False,self_thresh=False):
    '''
    Make a fancy dendrogram
    '''
    # Make the dendrogram
    if threshold == False:
        scipy.cluster.hierarchy.dendrogram(linkage,labels=names,orientation='right')
    else:
        scipy.cluster.hierarchy.dendrogram(linkage,labels=names, color_threshold=threshold,\
                                            orientation='right')

    # Color the names
    if name2color != False:
        ax = plt.gca()
        xlbls = ax.get_ymajorticklabels()
        for lbl in xlbls:
            lbl.set_color(name2color[lbl.get_text()])

    # Add the threshold
    if threshold:
        plt.axvline(x=threshold, c='k', linestyle='dotted')
    if self_thresh != False:
        plt.axvline(x=self_thresh, c='red', linestyle='dotted', lw=1)

    g = plt.gcf()
    return g


#!/usr/bin/env python

import sys
import subprocess
import matplotlib
import shutil
import seaborn
import numpy
import Bio
import pandas
import pysam

def humanbytes(B):
   '''
   Return the given bytes as a human friendly KB, MB, GB, or TB string
   https://stackoverflow.com/questions/12523586/python-format-size-application-converting-b-to-kb-mb-gb-tb/37423778
   '''
   B = float(B)
   KB = float(1024)
   MB = float(KB ** 2) # 1,048,576
   GB = float(KB ** 3) # 1,073,741,824
   TB = float(KB ** 4) # 1,099,511,627,776

   if B < KB:
      return '{0} {1}'.format(B,'Bytes' if 0 == B > 1 else 'Byte')
   elif KB <= B < MB:
      return '{0:.2f} KB'.format(B/KB)
   elif MB <= B < GB:
      return '{0:.2f} MB'.format(B/MB)
   elif GB <= B < TB:
      return '{0:.2f} GB'.format(B/GB)
   elif TB <= B:
      return '{0:.2f} TB'.format(B/TB)

def gen_dependency_report():
   """
   Return a string listing a number of dependencies and their versions
   """
   PROGRAMS = ['samtools', 'coverm']
   MODULES = [matplotlib, seaborn, numpy, Bio, pysam, pandas]

   ret_string = f'{"$"*10} DEPENDENCY REPORT {"$"*10}\n'
   ret_string += f'PYTHON\nRunning python v{sys.version}\n'

   ret_string += '\nPROGRAM DEPENDENCIES\n'
   for program in PROGRAMS:
      loc, works, version = find_program(program)
      works_message = {True: 'all good', False: '! NOT WORKING !'}[works]
      ret_string += f'{program:.<30} {works_message:10} (version={version}) (location = {loc})\n'.format(program, works_message, loc)

   ret_string += '\nPYTHON DEPENDENCIES\n'
   for module in MODULES:
      v = str(module.__version__)
      l = str(module.__file__)
      s = f"{l.split('/')[-2]:.<30} all good   (version={v}) (location = {l})\n"
      ret_string += s

   ret_string += f'{"$"*40}\n'

   return ret_string


def find_program(dep):
   """
   return location of progrgam, works = True/False (based on calling the help), and the version
   """
   # find the location of the program
   loc = shutil.which(dep)

   # make sure the help on the program works
   works = False
   v = 'na'
   if loc != None:
      try:
         v = subprocess.check_output([loc, '--version'], stderr=subprocess.STDOUT)
         v = str(v, 'utf-8').strip().split('\n')[0]
         works = True
      except:
         pass

   return loc, works, v
